#!/usr/bin/env python
#-*- coding: utf-8 -*-
#
# Copyright 2013-2019 European Commission (JRC);
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl
"""The command-line entry-point for using all functionality of wltp tool.

.. Seealso::
    :func:`main()`
"""

from __future__ import division, unicode_literals

import os, sys, collections, functools, re, glob, shutil, logging
from argparse import ArgumentTypeError
from collections import OrderedDict
import argparse
import pkg_resources as pkg
import ast
from distutils.spawn import find_executable
import json
from textwrap import dedent
from wltp import model, pandel, utils, __version__ as prog_ver
from wltp.pandel import JsonPointerException
from pandas.core.generic import NDFrame
import six

import jsonschema as jsons
import operator as ops
import pandas as pd


DEBUG   = False
PROG    = 'wltp'

DEFAULT_LOG_LEVEL   = logging.INFO
def _init_logging(loglevel, name='%s-cmd'%PROG, skip_root_level=False):
    logging.basicConfig(level=loglevel)

    rlog = logging.getLogger()
    if not skip_root_level:
        ## Force root-level, in case already configured otherwise.
        rlog.setLevel(loglevel)

    log = logging.getLogger(name)
    log.trace = lambda *args, **kws: log.log(0, *args, **kws)

    return log


log = _init_logging(DEFAULT_LOG_LEVEL)

def main(argv=None):
    """Calculates an engine-map by fitting data-points vectors, use --help for gettting help.

    REMARKS:
    --------
        * All string-values are case-sensitive.
        * Boolean string-values are case insensitive:
            False  : false, off, no,  == 0
            True   : true,  on,  yes, != 0
        * In KEY=VALUE pairs, the values are passed as string.  For other types,
          substitute '=' with:.
            +=     : integer
            *=     : float
            ?=     : boolean
            :=     : parsed as json
            @=     : parsed as python (with eval())

    EXAMPLES:
    ---------
    Assuming a 'vehicle.json' file like this:
        {
            "unladen_mass":1230,
            "test_mass":   1300,
            "v_max":    195,
            "p_rated":  100,
            "n_rated":  5450,
            "n_idle":   950,
            "n_min":    None, # Can be overriden by manufacturer.
            "gear_ratios":      [120.5, 75, 50, 43, 37, 32],
            "resistance_coeffs":[100, 0.5, 0.04],
        }

    then the following examples:

        ## Calculate and print fitted engine map's parameters
        #     for a petrol vehicle with the above engine-point's CSV-table:
        >> %(prog)s -I vehicle.csv file_model_path=/vehicle
        -I vehicle.csv file_frmt=SERIES model_path=/params header@=None \

        ## ...and if no header existed:
        >> %(prog)s -m /vehicle:=n_idle -I engine.csv header@=None

        ## Assume PME column contained normalized-Power in Watts,
        #    instead of P in kW:
        >> %(prog)s -m fuel=petrol -I engine.csv  -irenames X X 'Pnorm (w)'

        ## Read the same table above but without header-row and
        #    store results into Excel file, 1st sheet:
        >> %(prog)s -m fuel=petrol -I engine.csv --icolumns CM PME PMF -I engine_map.xlsx sheetname+=0

        ## Supply as inline-json more model-values required for columns [RPM, P, FC]
        #    read from <stdin> as json 2D-array of values (no headers).
        #    and store results in UTF-8 regardless of platform's default encoding:
        >> %(prog)s -m '/engine:={"fuel":"petrol", "stroke":15, "capacity":1359}' \\
                -I - file_frmt=JSON orient=values -c RPM P FC \\
                -O engine_map.txt encoding=UTF-8


    Now, if input vectors are in 2 separate files, the 1st, 'engine_1.xlsx',
    having 5 columns with different headers than expected, like this:
        OTHER1   OTHER2       N        "Fuel waste"   OTHER3
        0       -1            12       0.14           "some text"
        ...

    and the 2nd having 2 columns with no headers at all and
    the 1st column being 'Pnorm', then it, then use the following command:

        >> %(prog)s -O engine_map -m fuel=petrol \\
                -I=engine_1.xlsx sheetname+=0 \\
                -c X   X   N   'Fuel consumption'  X \\
                -r X   X   RPM 'FC(g/s)'           X \\
                -I=engine_2.csv header@=None \\
                -c Pnorm X
    """

    global log, DEBUG

    program_name    = os.path.basename(sys.argv[0])

    if argv is None:
        argv = sys.argv[1:]

    doc_lines       = main.__doc__.splitlines()
    desc            = doc_lines[0]
    epilog          = dedent('\n'.join(doc_lines[1:]))
    parser = build_args_parser(program_name, prog_ver, desc, epilog)

    opts = parser.parse_args(argv)

    try:
        DEBUG = bool(opts.debug)

        if (DEBUG or opts.verbose > 1):
            opts.strict = True

        if opts.verbose >= 2:
            level = 0
        elif opts.verbose >= 1:
            level = logging.DEBUG
        else:
            level = DEFAULT_LOG_LEVEL
        _init_logging(level, name=program_name)

        log.debug("Args: %s\n  +--Opts: %s", argv, opts)

        if opts.excel:
            copy_excel_template_files(opts.excel)
            return

#        if opts.excelrun:
#            files_copied = copy_excel_template_files(opts.excelrun)          #@UnusedVariable
#            xls_file = files_copied[0]
#
#            utils.open_file_with_os(xls_file)
#            return

        if opts.winmenus:
            add_windows_shortcuts_to_start_menu('winmenus')
            return

        opts = validate_file_opts(opts)

        infiles     = parse_many_file_args(opts.I, 'r', opts.irenames)
        log.info("Input-files: %s", infiles)

        outfiles    = parse_many_file_args(opts.O, 'w', None)
        log.info("Output-files: %s", outfiles)

    except (ValueError) as ex:
        if DEBUG:
            log.exception('Cmd-line parsing failed!')
        indent = len(program_name) * " "
        parser.exit(3, "%s: %s\n%s  for help use --help\n"%(program_name, ex, indent))

    ## Main program
    #
    try:
        additional_props = not opts.strict
        mdl = assemble_model(infiles, opts.m)
        log.info("Input Model(strict: %s): %s", opts.strict, utils.Lazy(lambda: model.json_dumps(mdl, 'to_string')))
        mdl = model.validate_model(mdl, additional_props)

        mdl = processor.run(opts, mdl)

        store_model_parts(mdl, outfiles)

    except jsons.ValidationError as ex:
        if DEBUG:
            log.error('Invalid input model!', exc_info=ex)
        indent = len(program_name) * " "
        parser.exit(4, "%s: Model validation failed due to: %s\n%s\n"%(program_name, ex, indent))

    except JsonPointerException as ex:
        if DEBUG:
            log.exception('Invalid model operation!')
        indent = len(program_name) * " "
        parser.exit(4, "%s: Model operation failed due to: %s\n%s\n"%(program_name, ex, indent))



def copy_excel_template_files(dest_dir=None):
    import pkg_resources as pkg

    if not dest_dir == None:
        dest_dir = os.getcwd()
    else:
        dest_dir = os.path.abspath(dest_dir)

    try:
        os.mkdir(dest_dir)
        log.info('Created destination-directory(%s).', dest_dir)
    except:
        pass ## Might already exist

    files_to_copy = ['excel\WltpExcelRunner.xlsm', 'excel\WltpExcelRunner.py']
    files_to_copy = [pkg.resource_filename('wltp', f) for f in files_to_copy] #@UndefinedVariable
    files_copied = []
    for src_fname in files_to_copy:
        dest_fname = os.path.basename(src_fname)
        fname_genor = utils.generate_filenames(os.path.join(dest_dir, dest_fname))
        dest_fname = next(fname_genor)
        while os.path.exists(dest_fname):
            dest_fname = next(fname_genor)

        log.info('Copying `xlwings` template-file: %s --> %s', src_fname, dest_fname)
        shutil.copy(src_fname, dest_fname)
        files_copied.append(dest_fname)

    return files_copied


def add_windows_shortcuts_to_start_menu(my_option):
    if sys.platform != 'win32':
        exit('This options can run only under *Windows*!')
    my_cmd_name = PROG
    my_cmd_path = find_executable(my_cmd_name)
    if not my_cmd_path:
        exit("Please properly install the project before running the `--%s` command-option!" % my_option)

    win_menu_group = 'Python Wltp Calculator'

    wshell = utils.win_wshell()
    startMenu_dir   = utils.win_folder(wshell, "StartMenu")
    myDocs_dir      = utils.win_folder(wshell, "MyDocuments")
    docs_url        = 'http://wltp.readthedocs.org/'
    prog_dir        = os.path.join(myDocs_dir, '%s_%s'%(PROG, prog_ver))
    shcuts = OrderedDict([
        ("Create new Wltp ExcelRunner files.lnk", {
            'target_path':  'cmd',
            'target_args':  '/K wltp --excelrun',
            'wdir':         prog_dir,
            'desc':         'Copy `xlwings` excel & python template files into `MyDocuments` and open the Excel-file, so you can run a batch of experiments.',
            'icon_path':    pkg.resource_filename('wltp.excel', 'ExcelPython.ico'),        #@UndefinedVariable
        }),
        ("Wltp Documentation site.url", {
            'target_path':  docs_url,
        }),
    ])


    try:
        os.makedirs(prog_dir, exist_ok=True)
    except Exception as ex:
        log.exception('Failed creating Program-folder(%s) due to: %s', prog_dir, ex)
        exit(-5)

    ## Copy Demos.
    #
    demo_dir = pkg.resource_filename('wltp', 'test')                     #@UndefinedVariable
    demo_files =['*.csv', '*.xls?', '*.bat', 'engine.py', 'cmdline_test.py']
    for g in demo_files:
        for src_f in glob.glob(os.path.join(demo_dir, g)):
            f = os.path.basename(src_f)
            dest_f = os.path.join(prog_dir, f)
            if os.path.exists(dest_f):
                log.trace('Removing previous entry(%s) from existing Program-folder(%s).', f, prog_dir)
                try:
                    os.unlink(dest_f)
                except Exception as ex:
                    log.error('Cannot clear existing item(%s) from Program-folder(%s) due to: %s', dest_f, prog_dir, ex)

            try:
                shutil.copy(src_f, dest_f)
            except Exception as ex:
                log.exception('Failed copying item(%s) in program folder(%s): %s', src_f, prog_dir, ex)

    ## Create a fresh-new Menu-group.
    #
    group_path = os.path.join(startMenu_dir, win_menu_group)
    if os.path.exists(group_path):
        log.trace('Removing all entries from existing StartMenu-group(%s).', win_menu_group)
        for f in glob.glob(os.path.join(group_path, '*')):
            try:
                os.unlink(f)
            except Exception as ex:
                log.warning('Minor failure while removing previous StartMenu-item(%s): %s', f, ex)

    try:
        os.makedirs(group_path, exist_ok=True)
    except Exception as ex:
        log.exception('Failed creating StarMenu-group(%s) due to: %s', group_path, ex)
        exit(-6)

    for name, shcut in shcuts.items():
        path = os.path.join(group_path, name)
        log.info('Creating StartMenu-item: %s', path)
        try:
            utils.win_create_shortcut(wshell, path, **shcut)
        except Exception as ex:
            log.exception('Failed creating item(%s) in StartMenu-group(%s): %s', path, win_menu_group , ex)


## The value of file_frmt=VALUE to decide which
#    pandas.read_XXX() and write_XXX() methods to use.
#
#_io_file_modes = {'r':0, 'rb':0, 'w':1, 'wb':1, 'a':1, 'ab':1}
_io_file_modes = {'r':0, 'w':1}
_read_clipboard_methods = (pd.read_clipboard, 'to_clipboard')
_default_pandas_format  = 'AUTO'
_pandas_formats = collections.OrderedDict([
    ('AUTO', None),
    ('CSV', (pd.read_csv, 'to_csv')),
    ('TXT', (pd.read_csv, 'to_csv')),
    ('XLS', (pd.read_excel, 'to_excel')),
    ('JSON', (pd.read_json, 'to_json')),
    ('SERIES', (pd.Series.from_csv, 'to_json')),
])
_known_file_exts = {
    'XLSX':'XLS'
}
def get_file_format_from_extension(fname):
    ext = os.path.splitext(fname)[1]
    if (ext):
        ext = ext.upper()[1:]
        if (ext in _known_file_exts):
            return _known_file_exts[ext]
        if (ext in _pandas_formats):
            return ext

    return None


_default_df_path            = ('/engine_points', '/engine_map')
_default_out_file_append    = False
## When option `-m MODEL_PATH=VALUE` contains a relative path,
# the following is preppended:
_default_model_overridde_path = '/engine/'


_value_parsers = {
    '+': int,
    '*': float,
    '?': utils.str2bool,
    ':': json.loads,
    '@': eval,
    #'@': ast.literal_eval ## best-effort security: http://stackoverflow.com/questions/3513292/python-make-eval-safe
}


_key_value_regex = re.compile(r'^\s*([/_A-Za-z][\w/\.]*)\s*([+*?:@]?)=\s*(.*?)\s*$')
def parse_key_value_pair(arg):
    """Argument-type for syntax like: KEY [+*?:]= VALUE."""

    m = _key_value_regex.match(arg)
    if m:
        (key, type_sym, value) = m.groups()
        if type_sym:
            try:
                value   = _value_parsers[type_sym](value)
            except Exception as ex:
                raise six.reraise(ArgumentTypeError, ("Failed parsing key(%s)%s=VALUE(%s) due to: %s" %(key, type_sym, value, ex)), ex.__traceback__) ## Python-2 :-(

        return [key, value]
    else:
        raise argparse.ArgumentTypeError("Not a KEY=VALUE syntax: %s"%arg)


def parse_column_specifier(arg):
    """Argument-type for --icolumns, syntaxed like: COL_NAME [(UNITS)]."""

    res = pandel.parse_value_with_units(arg)
    if res:
        return res

    raise argparse.ArgumentTypeError("Not a COLUMN_SPEC syntax: %s"%arg)


def validate_file_opts(opts):
    ## Check number of input-files <--> related-opts
    #
    dopts = vars(opts)

    if (not opts.I):
        n_infiles = 1
    else:
        n_infiles = len(opts.I)
    rel_opts = ['icolumns', 'irenames']
    for ropt in rel_opts:
        opt_val = dopts[ropt]
        if (opt_val):
            n_ropt = len(opt_val)
            if( n_ropt > 1 and n_ropt != n_infiles):
                raise argparse.ArgumentTypeError("Number of --%s(%i) mismatches number of -I(%i)!"%(ropt, n_ropt, n_infiles))


    return opts


FileSpec = collections.namedtuple('FileSpec', ('io_method', 'fname', 'file', 'frmt', 'path', 'append', 'renames', 'kws'))

def parse_many_file_args(many_file_args, filemode, col_renames=None):
    io_file_indx = _io_file_modes[filemode]

    def parse_file_args(n, fname, *kv_args):
        frmt    = _default_pandas_format
        path    = _default_df_path[io_file_indx]
        append  = _default_out_file_append

        kv_pairs = [parse_key_value_pair(kv) for kv in kv_args]
        pandas_kws = dict(kv_pairs)

        if ('file_frmt' in pandas_kws):
            frmt = pandas_kws.pop('file_frmt')
            if (frmt not in _pandas_formats):
                raise argparse.ArgumentTypeError("Unsupported pandas file_frmt: %s\n  Set 'file_frmt=XXX' to one of %s" % (frmt, list(_pandas_formats.keys())[1:]))


        if (frmt == _default_pandas_format):
            if ('-' == fname or fname == '+'):
                frmt = 'CSV'
            else:
                frmt = get_file_format_from_extension(fname)
            if (not frmt):
                raise argparse.ArgumentTypeError("File(%s) has unknown extension, file_frmt is required! \n  Set 'file_frmt=XXX' to one of %s" % (fname, list(_pandas_formats.keys())[1:]))


        if (fname == '+'):
            method = _read_clipboard_methods[io_file_indx]
            file = ''
        else:
            methods = _pandas_formats[frmt]
            assert isinstance(methods, tuple), methods
            method = methods[io_file_indx]

            if (method == pd.read_excel):
                file = fname
            else:
                file = argparse.FileType(filemode)(fname)

        try:
            path = pandas_kws.pop('model_path')
            if (len(path) > 0 and not path.startswith('/')):
                raise argparse.ArgumentTypeError("Only absolute model-paths (those starting with '/') are supported: %s" % (path))
        except KeyError:
            pass

        try:
            append = pandas_kws.pop('file_append')
            append = utils.str2bool(append)
        except KeyError:
            pass

        ## Here we apply a single --irenames to all input-files.
        #
        if col_renames is None:
            renames = None
        elif len(col_renames) == 1:
            renames = col_renames[0]
        else:
            renames = col_renames[n]
        return FileSpec(method, fname, file, frmt, path, append, renames, pandas_kws)

    if not many_file_args:
        return [] # FIXME: Why enumeration on None does not work?
    return [parse_file_args(n, *file_args) for (n, file_args) in enumerate(many_file_args)]


def load_file_as_df(filespec):
# FileSpec(io_method, fname, file, frmt, path, append, kws)
    method = filespec.io_method
    log.debug('Reading file with: pandas.%s(%s, %s)', method.__name__, filespec.fname, filespec.kws)
    if filespec.file is None:       ## ie. when reading CLIPBOARD
        dfin = method(**filespec.kws)
    else:
        dfin = method(filespec.file, **filespec.kws)

    if (filespec.renames):
        old_cols = dfin.columns
        new_cols = [old if new['name'] == '_' else new['name'] for (old, new) in zip(old_cols, filespec.renames)]

        ## If length of rename-columns differ from DF's.
        #    columns at the end are left unchanged,
        #    but will scream if given more renames!
        #
        new_cols += list(old_cols[len(new_cols):])
        dfin.columns = new_cols


    dfin = dfin.convert_objects(convert_numeric=True)

    return dfin


def load_model_part(mdl, filespec):
    dfin = load_file_as_df(filespec)
    log.debug("  +-input-file(%s):\n%s", filespec.fname, dfin.head())
    if filespec.path:
        pandel.set_jsonpointer(mdl, filespec.path, dfin)
    else:
        mdl = dfin
    return mdl


def assemble_model(infiles, model_overrides):

    mdl = model._get_model_base()

    for filespec in infiles:
        try:
            mdl = load_model_part(mdl, filespec)
        except Exception as ex:
            raise six.reraise(Exception, ("Failed reading %s due to: %s" %(filespec, ex)), ex.__traceback__) ## Python-2 :-(

    if (model_overrides):
        model_overrides = functools.reduce(lambda x,y: x+y, model_overrides) # join all -m
        for (json_path, value) in model_overrides:
            try:
                if (not json_path.startswith('/')):
                    json_path = _default_model_overridde_path + json_path
                pandel.set_jsonpointer(mdl, json_path, value)
            except Exception as ex:
                raise six.reraise(Exception, ("Failed setting model-value(%s) due to: %s" %(json_path, value, ex)), ex.__traceback__) ## Python-2 :-(

    return mdl


def store_part_as_df(filespec, part):
    '''If part is Pandas, store it as it is, else, store it as json recursively.

        :param FileSpec filespec: named_tuple
        :param part: what to store, originating from model(filespec.path))
    '''

    if isinstance(part, NDFrame):
        log.debug('Writing file with: pandas.%s(%s, %s)', filespec.io_method, filespec.fname, filespec.kws)
        if filespec.file is None:       ## ie. when reading CLIPBOARD
            method = ops.methodcaller(filespec.io_method, **filespec.kws)
        else:
            method = ops.methodcaller(filespec.io_method, filespec.file, **filespec.kws)
        method(part)
    else:
        model.json_dump(part, filespec.file, pd_method=None, **filespec.kws)


def store_model_parts(mdl, outfiles):
    for filespec in outfiles:
        try:
            try:
                part = pandel.resolve_jsonpointer(mdl, filespec.path)
            except JsonPointerException:
                log.warning('Nothing found at model(%s) to write to file(%s).', filespec.path, filespec.fname)
            else:
                store_part_as_df(filespec, part)
        except Exception as ex:
            raise six.reraise(Exception, ("Failed storing %s due to: %s" %(filespec, ex)), ex.__traceback__) ## Python-2 :-(

class RawTextHelpFormatter(argparse.RawDescriptionHelpFormatter):
    """Help message formatter which retains formatting of all help text.

    Only the name of this class is considered a public API. All the methods
    provided by the class are considered an implementation detail.
    """

    def _split_lines(self, text, width):
        return text.splitlines()


def build_args_parser(program_name, version, desc, epilog):
    version_string  = '%s' % version

    parser = argparse.ArgumentParser(prog=program_name, description=desc, epilog=epilog, add_help=False,
                                     formatter_class=RawTextHelpFormatter)


    grp_io = parser.add_argument_group('Input/Output', 'Options controlling reading/writting of file(s) and for specifying model values.')
    grp_io.add_argument('-I', help=dedent("""\
            import file(s) into the model utilizing pandas-IO methods, see
                http://pandas.pydata.org/pandas-docs/stable/io.html
            Default: %(default)s]
            * The syntax of this option is like this:
                    FILENAME [KEY=VALUE ...]]
            * The FILENAME can be '-' to designate <stdin> or '+' to designate CLIPBOARD.
            * Any KEY-VALUE pairs pass directly to pandas.read_XXX() options,
              except from the following keys, which are consumed before reaching pandas:
                ** file_frmt = [ AUTO | CSV | TXT | XLS | JSON | SERIES ]
                  selects which pandas.read_XXX() method to use:
                    *** AUTO: the format is deduced from the filename's extension (ie Excel files).
                    *** JSON: different sub-formats are selected through the 'orient' keyword
                      of Pandas specified with a key-value pair
                      (see: http://pandas.pydata.org/pandas-docs/dev/generated/pandas.io.json.read_json.html).
                    *** SERIES: uses `pd.Series.from_csv()`.
                    *** Defaults to AUTO, unless reading <stdin> or <clipboard>, which then is CSV.
                ** model_path = MODEL_PATH
                  specifies the destination (or source) of the dataframe within the model
                  as json-pointer path (see -m option).
                  If many input-files have the same --model_path, the dataframes are
                  concatenated horizontally, therefore the number of rows (excluding header)
                  for all those data-files must be equal.
            * When multiple input-files given, the number of --icolumns and --irenames options,
              must either match them, be 1 (meaning, use them for all files), or be totally absent
              (meaning, use defaults for all files).
            * see REMARKS at the bottom regarding the parsing of KEY-VAULE pairs. """),
                        action='append', nargs='+',
                        #default=[('- file_frmt=%s model_path=%s'%('CSV', _default_df_dest)).split()],
                        metavar='ARG')
    grp_io.add_argument('-c', '--icolumns', help=dedent("""\
            describes the columns-contents of input file(s) along with their units (see --I).
            It must be followed either by an integer denoting the index of the header-row
            within the tabular data, or by a list of column-names specifications,
            obeying the following syntax:
                COL_NAME [(UNITS)]
            Accepted quantities and their default units are grouped in 3+1 quantity-types and
            for each file exactly one from each of the 3 first categories must be present:
            1. engine-speed:
                RPM      (rad/min)
                RPMnorm  (rad/min)  : normalized against RPMnorm * RPM_IDLE + (RPM_RATED - RPM_IDLE)
                Omega    (rad/sec)
                CM       (m/sec)    : Mean Piston speed
            2. work-capability:
                P        (kW)
                Pnorm    (kW)       : normalized against P_MAX
                T        (Nm)
                PME      (bar)
            3. fuel-consumption:
                FC       (g/h)
                FCnorm   (g/h)      : normalized against P_MAX
                PMF      (bar)
            4. Irellevant column:
                X
            Default when files include headers is 0 (1st row), otherwise it is 'RPM,P,FC'."""),
                        action='append', nargs='+',
                        type=parse_column_specifier, metavar='COLUMN_SPEC')
    grp_io.add_argument('-r', '--irenames', help=dedent("""\
            renames the columns of input-file(s)  (see --I).
            It must be followed by a list of column-names specifications like --icolumns,
            but without accepting integers.
            The number of renamed-columns for each input-file must be equal or less
            than those in the --icolumns for the respective inpute-file.
            Use '_' for columns to be left intact."""),
                        action='append', nargs='*',
                        type=parse_column_specifier, metavar='COLUMN_SPEC')
    grp_io.add_argument('-m', help=dedent("""\
            override a model value.
            * The MODEL_PATH is a json-pointer absolute or relative path, see:
                    https://python-json-pointer.readthedocs.org/en/latest/tutorial.html
              Relative paths are resolved against '/engine', for instance:
                    -Mrpm_idle=850   -M/engine/p_max=660
              would set the following model's property:
                    {
                      "engine": {
                          "rpm_idle": 850,
                          "p_max": 660,
                          ...
                      }
                    }
            * Any values from -m options are applied AFTER any files read by -I option.
            * See REMARKS below for the parsing of KEY-VAULE pairs.
            """),
                        action='append', nargs='+',
                        type=parse_key_value_pair, metavar='MODEL_PATH=VALUE')
    grp_io.add_argument('--strict', help=dedent("""\
            validate model more strictly, ie no additional-properties allowed.
            [default: %(default)s]"""),
            default=False, type=utils.str2bool,
            metavar='[TRUE | FALSE]')
    grp_io.add_argument('-M', help=dedent("""\
            get help description for the specfied model path.
            If no path specified, gets the default model-base. """),
                        action='append', nargs='*', metavar='MODEL_PATH')


    grp_io.add_argument('-O', help=dedent("""\
            specifies output-file(s) to write model-portions into after calculations.
            The syntax is indentical to -I, with these differences:
            * Instead of <stdin>, <stdout> and write_XXX() methods are used wherever.
            * One extra key-value pair:
            ** file_append = [ TRUE | FALSE ]
                    specify whether to augment pre-existing files, or overwrite them.
            * Default: - file_frmt=CSV model_path=/engine_map """),
                        action='append', nargs='+',
                        #default=[('- file_frmt=%s model_path=%s file_append=%s'%('CSV', _default_df_path[1],  _default_out_file_append)).split()],
                        metavar='ARG')


    xlusive_group = parser.add_mutually_exclusive_group()
    xlusive_group.add_argument('--excel', help="copy `xlwings` excel & python template files into DESTPATH or current-working dir, to run a batch of experiments",
        nargs='?', const=None, metavar='DESTPATH')
#    xlusive_group.add_argument('--excelrun', help="Copy `xlwings` excel & python template files into USERDIR and open Excel-file, to run a batch of experiments",
#        nargs='?', const=os.getcwd(), metavar='DESTPATH')
    xlusive_group.add_argument('--winmenus', help="Adds shortcuts into Windows StartMenu.", action='store_true')

    grp_various = parser.add_argument_group('Various', 'Options controlling various other aspects.')
    grp_various.add_argument('-d', "--debug", action="store_true", help=dedent("""\
            set debug-mode with various checks and error-traces
            Suggested combining with --verbose counter-flag.
            Implies --strict true
            [default: %(default)s] """),
                        default=False)
    grp_various.add_argument('-v', "--verbose", action="count", default=0, help="set verbosity level [default: %(default)s]")
    grp_various.add_argument("--version", action="version", version=version_string, help="prints version identifier of the program")
    grp_various.add_argument("--help", action="help", help='show this help message and exit')

    return parser



if __name__ == "__main__":
    main()



#!/usr/bin/env python
#
"""
Script to bump, commit and tag new versions.

USAGE:
    bumpver [-n] [-f] [-c] [-t <message>]  [<new-ver>]

OPTIONS:
  -f, --force       Bump (and optionally) commit even if version is the same.
  -n, --dry-run     Do not write files - just pretend.
  -c, --commit      Commit afterwards.
  -t, --tag=<msg>   Adds a signed tag with the given message (commit implied).

Without <new-ver> prints version extracted from current file.

"""

import os.path as osp
import sys
import re
import functools as fnt

import docopt


my_dir = osp.dirname(__file__)

VFILE = osp.join(my_dir, '..', 'wltp', '_version.py')
VFILE_regex_v = re.compile(r'__version__ = "([^"]+)"')
VFILE_regex_d = re.compile(r'__updated__ = "([^"]+)"')

RFILE = osp.join(my_dir, '..', 'README.rst')


class CmdException(Exception):
    pass


@fnt.lru_cache()
def read_txtfile(fpath):
    with open(fpath, 'rt', encoding='utf-8') as fp:
        return fp.read()


def extract_file_regexes(fpath, regexes):
    """
    :param regexes:
        A sequence of regexes to "search", having a single capturing-group.
    :return:
        One groups per regex, or raise if any regex did not match.
    """
    txt = read_txtfile(fpath)
    matches = [regex.search(txt) for regex in regexes]

    if not all(matches):
        raise CmdException("Failed extracting current version: "
                           "\n  ver: %s\n  date: %s" % tuple(matches))

    return [m.group(1) for m in matches]


def replace_substrings(files, subst_pairs):
    for fpath in files:
        txt = read_txtfile(fpath)

        replacements = []
        for old, new in subst_pairs:
            replacements.append((old, new, txt.count(old)))
            txt = txt.replace(old, new)

        yield (txt, fpath, replacements)


def format_syscmd(cmd):
    if isinstance(cmd, (list, tuple)):
        cmd = ' '.join('"%s"' % s if ' ' in s else s
                       for s in cmd)
    else:
        assert isinstance(cmd, str), cmd

    return cmd


def strip_ver2_commonprefix(ver1, ver2):
    cprefix = osp.commonprefix([ver1, ver2])
    if cprefix:
        striplen = cprefix.rfind('.')
        if striplen > 0:
            striplen += 1
        else:
            striplen = len(cprefix)
        ver2 = ver2[striplen:]

    return ver2


def run_testcases():
    import unittest

    ## Needed to import TCs AFTER grafting of version.
    #
    sys.path.append(osp.normpath(osp.join(my_dir, '..')))
    import wltp.test.test_docs as test_docs  # @IgnorePep8

    suite = unittest.TestLoader().loadTestsFromModule(test_docs)
    res = unittest.TextTestRunner(failfast=True).run(suite)

    if not res.wasSuccessful():
        raise CmdException("Doc TCs failed, probably version-bumping has failed!")


def exec_cmd(cmd):
    import subprocess as sbp

    err = sbp.call(cmd, stderr=sbp.STDOUT)
    if err:
        raise CmdException("Failed(%i) on: %s" % (err, format_syscmd(cmd)))


def do_commit(new_ver, old_ver, dry_run, ver_files):
    import pathlib

    new_ver = strip_ver2_commonprefix(old_ver, new_ver)
    cmt_msg = 'chore(ver): bump %s-->%s' % (old_ver, new_ver)

    ver_files = [pathlib.Path(f).as_posix() for f in ver_files]
    commands = [
        ['git', 'add'] + ver_files,
        ['git', 'commit', '-m', cmt_msg]
    ]

    for cmd in commands:
        cmd_str = format_syscmd(cmd)
        if dry_run:
            yield "DRYRUN: %s" % cmd_str
        else:
            yield "EXEC: %s" % cmd_str
            exec_cmd(cmd)


def do_tag(tag, tag_msg, dry_run):
    cmd = ['git', 'tag', tag, '-s', '-m', tag_msg]
    cmd_str = format_syscmd(cmd)
    if dry_run:
        yield "DRYRUN: %s" % cmd_str
    else:
        yield "EXEC: %s" % cmd_str
        exec_cmd(cmd)


def bumpver(new_ver, dry_run=False, force=False, tag_after_commit=None):
    """
    :param tag_after_commit:
        if true, do `git commit`, if string, also `git tag` with that as msg.
    """
    regexes = [VFILE_regex_v, VFILE_regex_d]
    old_ver, old_date = extract_file_regexes(VFILE, regexes)

    if not new_ver:
        yield old_ver
        yield old_date
    else:
        if new_ver == old_ver:
            msg = "Version '%s'already bumped"
            if force:
                msg += ", but --force  effected."
                yield msg % new_ver
            else:
                msg += "!\n Use of --force recommended."
                raise CmdException(msg % new_ver)

        from datetime import datetime

        new_date = datetime.now().strftime('%Y-%m-%d %H:%M:%S%z')

        ver_files = [osp.normpath(f) for f in [VFILE, RFILE]]
        subst_pairs = [(old_ver, new_ver), (old_date, new_date)]

        for repl in replace_substrings(ver_files, subst_pairs):
            new_txt, fpath, replacements = repl

            if not dry_run:
                with open(fpath, 'wt', encoding='utf-8') as fp:
                    fp.write(new_txt)

            yield '%s: ' % fpath
            for old, new, nrepl in replacements:
                yield '  %i x (%24s --> %s)' % (nrepl, old, new)

        yield "...now launching DocTCs..."
        run_testcases()

        if tag_after_commit is not None:
            yield from do_commit(new_ver, old_ver, dry_run, ver_files)

            if isinstance(tag_after_commit, str):
                tag = 'v%s' % new_ver
                yield from do_tag(tag, tag_after_commit, dry_run)


def main(*args):
    opts = docopt.docopt(__doc__, argv=args)

    new_ver = opts['<new-ver>']
    dry_run = opts['--dry-run']
    force = opts['--force']

    commit = opts['--commit']
    tag = opts['--tag']
    if tag:
        tag_after_commit = tag
    elif commit:
        tag_after_commit = True
    else:
        tag_after_commit = None

    try:
        for i in bumpver(new_ver, dry_run, force, tag_after_commit):
            print(i)
    except CmdException as ex:
        sys.exit(str(ex))
    except Exception as ex:
        raise ex


if __name__ == '__main__':
    main(*sys.argv[1:])



#!/usr/bin/env python
#-*- coding: utf-8 -*-
#
# Copyright 2013-2019 European Commission (JRC);
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl
'''
Converts to CSVs and consolidate input & results tables from Heinz' ms-access db.

:input inp_excel:   one excel-file with input vehicles, exported from the ``vehicle_info`` db-table
:input veh_excel:   many excel-files exported from ``A gearshift_table_check_result_final`` query,
                     one for each row in the above file
:output inp_csv:   a csv from the input-excel, updated with same-valued(steady) columns
:output veh_csv:   multiple CSVs from the input-excels, without the same-valued(steady) columns (moved to inp_vehs)

.. Note: Requires excel-files to be present that were produced from ``A gearshift_table_check_result_final``
ms-access query.
'''

from __future__ import division, print_function, unicode_literals

import glob, os, os.path as path
import re

from openpyxl.exceptions import IllegalCharacterError

import numpy as np, pandas as pd


def heinz_db_result_table_to_csv(heinz_results_glob, inp_vehs_df, cols_to_move, cols_to_rename, cols_to_drop, index_col):

    for heinz_fname in heinz_results_glob:
        print("%s..."%heinz_fname, end='')

        try:
            m = re.match('(\d+)', heinz_fname)
            assert m, heinz_fname
            veh = int(m.groups()[0])

            assert veh in inp_vehs_df.index, "Heinz-out veh(%s) not in inp-vehs!"%veh

            heinz_out_df = pd.read_excel(heinz_fname, 0)

            ## Check veh's fname matches content.
            assert veh == heinz_out_df.ix[0, 'vehicle_no'], "heinz-out[0, 0](%s) != fname-veh(%s)!"%(heinz_out_df.iloc[0, 0], veh)

            ## Check same-valued columns are indeed so.
            #
            steady_cols = cols_to_move + [left for (left, _) in cols_to_rename]
            for col in steady_cols:
                cval = heinz_out_df.ix[0, col]
                assert (heinz_out_df[col] == cval).all(), "move-col(%s): %s != for all-heinz-out(%s)!" %(col, cval, heinz_out_df[col])


            ## Rename same-valued columns with clashing names into input-matrix.
            #
            for (col, ncol) in cols_to_rename:
                cval = heinz_out_df.ix[0, col]
                if ncol in inp_vehs_df.columns:
                    ## Check if overwriting other value.
                    #
                    dval = inp_vehs_df.ix[veh, ncol]
                    if not pd.isnull(dval) and heinz_out_df[col].dtype == inp_vehs_df[ncol].dtype and cval != dval:
                        print("Warn: rename-col(%s->%s): %s != inp-vehs(%s)!" %(col, ncol, cval, dval))
                inp_vehs_df.ix[veh, ncol] = cval

            ## Move same-valued columns into input-matrix.
            #
            for col in cols_to_move:
                cval = heinz_out_df.ix[0, col]
                if col in inp_vehs_df.columns:
                    ## Check if overwriting other value.
                    #
                    dval = inp_vehs_df.ix[veh, col]
                    if not pd.isnull(dval) and heinz_out_df[col].dtype == inp_vehs_df[col].dtype and cval != dval:
                        print("WARN: move-col(%s): %s != inp-vehs(%s)!" %(col, cval, dval))
                inp_vehs_df.ix[veh, col] = cval


            heinz_out_df.index = heinz_out_df[index_col]
            heinz_out_df.index.name = index_col

            heinz_out_df = heinz_out_df.drop(steady_cols + cols_to_drop + [index_col], axis=1)

            (fn, _) = os.path.splitext(heinz_fname)
            outfname = 'heinz-{:04}{}'.format(veh, ".csv")

            heinz_out_df.to_csv(outfname, encoding='UTF-8')
            print("  Stored heiz-out(%s)" %outfname)
        except Exception as ex:
            print("Error: ", ex)

    return inp_vehs_df





def post_proc_heinz_results():
    """
    Moves columns with repeatitious data into a master-vehicles csv listing all vehicle-parameters.
    """

    inp_vehs_xcls            = 'vehicle_info.xlsx'
    inp_vehs_fname          = 'wltp_db_vehicles.csv'
    inp_vehs_2_fname        = 'wltp_db_vehicles2.csv'
    heinz_results_wildcard  = '*-*.xls'
    index_col               = 't'
    ## Move same-valued columns from heinz-output's --> input-matrix.
    cols_to_move            = [
        'Power_curve_no', 'safety_margin_Pwot', 'additional_margin_at_idling', 'actual_safety_margin',
        'safety_margin_v_max', 'downscale_percentage', 'look_ahead_time', 'vehicle_no', 'kerb_mass', 'test_mass',
        'rated_power', 'rated_speed', 'idling_speed', 'n_min_drive', 'n_min_2', 'vehicle_class', 'n_norm_max', 'no_of_gears',
    ]
    ## Copy renamed same-valued columns from heinz-output's --> input-matrix.
    cols_to_rename          = [
        ('v_max',  'v_max_2'),
        ('f0', 'f0_real'), ('f1', 'f1_real'), ('f2', 'f2_real' )
    ]
    ## Columns to remove from heinz-output's.
    cols_to_drop            = [ 'f0', 'f1', 'f2',  ]


    # (inp_vehs_2_fname, ext) = path.splitext(inp_vehs_fname)
    #inp_vehs_2_fname = '%s-2%s'%(inp_vehs_2_fname, ext)

    heinz_results_glob = glob.glob(heinz_results_wildcard)

    dfin = pd.read_excel(inp_vehs_xcls, 0, index_col=0)
    #dfin = pd.read_csv(inp_vehs_fname, index_col=0, encoding='UTF-8')

    dfin = heinz_db_result_table_to_csv(heinz_results_glob, dfin, cols_to_move, cols_to_rename, cols_to_drop, index_col)

    ## Remove vehs without results (no f0)
    dfin=dfin[~dfin.f0_real.isnull()]

    dfin.to_csv(inp_vehs_2_fname, encoding='UTF-8')
    print("Stored modified inp-vehs(%s)" %inp_vehs_2_fname)


def reindex_db_vehicles():
    heinz_classes_excel = 'gearshift_description_development_DB_mod2.xlsx'
    out_vehs_fname = 'wltp_db_vehicles.xls'

    dfin1 = pd.read_excel(heinz_classes_excel, 0, index_col=None, skiprows=4, parse_cols='C:Z')
    dfin2 = pd.read_excel(heinz_classes_excel, 1, index_col=None, skiprows=4, parse_cols='C:Z')
    dfin3 = pd.read_excel(heinz_classes_excel, 2, index_col=None, skiprows=4, parse_cols='C:Z')

    dfout = pd.concat((dfin1, dfin2, dfin3, ), ignore_index=True)

    ## Move old possible-repetetitious veh_no into comments.
    #
    dfout['description1'] = dfout['vehicle_no'].apply(lambda x: '%s: '%x) + dfout['description1'].apply(lambda x: str(x))
    dfout['vehicle_no'] = dfout.reset_index().index

    dfout = dfout.rename(columns={
        'category': 'ID_cat',
        'description1': 'Comment',
    })
    dfout['eng_id'] = dfout['IDengine']

    ## New cols.
    #
#     dfout['ndv_7'] = np.nan
    dfout['user_def_driv_res_coeff'] = 0
    dfout['user_def_power_curve'] = 0

    dfout = dfout.reindex(columns = [
        'vehicle_no', 'rated_power', 'kerb_mass', 'rated_speed', 'idling_speed', 'test_mass', 'no_of_gears',
        'ndv_1', 'ndv_2', 'ndv_3', 'ndv_4', 'ndv_5', 'ndv_6', 'ndv_7',
        'ID_cat', 'user_def_driv_res_coeff', 'user_def_power_curve', 'f0', 'f1', 'f2', 'Comment',
        'pmr_km', 'v_max', 'g_v_max', 'n_norm_vmax', 'n_vmax',  # Should have dropped them.
    ])

    NGs = dfout.loc[:, 'ndv_1':'ndv_7']
    NGs[NGs == 0] = np.NaN
    dfout.loc[:, 'ndv_1':'ndv_7'] = NGs

    dfout.to_excel(out_vehs_fname, index=False)


if __name__ == '__main__':
    _mydir = path.dirname(__file__)
    os.chdir(path.join(_mydir, '../wltp/test/wltp_db'))

    post_proc_heinz_results() #TODO: have top re-proc heinz-files.
    #reindex_db_vehicles()



#!/usr/bin/env python
#-*- coding: utf-8 -*-
#
# Copyright 2013-2019 European Commission (JRC);
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl


if __name__ == '__main__':
    import gears_n_arrows as p
    from matplotlib import pyplot as plt

    p.plot(1)
    plt.show()


#!/usr/bin/python3
"""
Sed-file to clean-up Sphinx-only constructs (ie from README.rst),
so that *PyPi* can format it properly.

To check for remaining errors, install ``sphinx`` and run::

        python setup.py --long-description | sed -file 'this_file.sed' | rst2html.py  --halt=warning

"""

from __future__ import division, print_function, unicode_literals

import re
import sys, io


def clean_sphinx_only_markup(file_inp=None, file_out=None):
    """
    :param file_inp:     a `filename` or ``sys.stdin``?
    :param file_out:     a `filename` or ``sys.stdout`?`

    """
    substs = [
        ## Selected Sphinx-only Roles.
        #
        (r':abbr:`([^`]+)`',        r'\1'),
        (r':ref:`([^`]+)`',         r'`\1`_'),
        (r':term:`([^`]+)`',        r'**\1**'),
        (r':dfn:`([^`]+)`',         r'**\1**'),
        (r':(samp|guilabel|menuselection):`([^`]+)`',        r'``\2``'),


        ## Sphinx-only roles:
        #        :foo:`bar`   --> foo(``bar``)
        #        :a:foo:`bar` XXX afoo(``bar``)
        #
        #(r'(:(\w+))?:(\w+):`([^`]*)`', r'\2\3(``\4``)'),
        (r':(\w+):`([^`]*)`', r'\1(``\2``)'),


        ## Sphinx-only Directives.
        #
        (r'\.\. +doctest',           r'code-block'),
        (r'\.\. +plot::',            r'.. '),
        (r'\.\. +seealso',           r'info'),
        (r'\.\. +glossary',          r'rubric'),
        (r'\.\. +figure::',          r'.. '),


        ## Other
        #
        (r'\|version\|',              r'x.x.x'),
    ]

    regex_subs = [ (re.compile(regex, re.IGNORECASE), sub) for (regex, sub) in substs ]

    def clean_line(line):
        try:
            for (regex, sub) in regex_subs:
                line = regex.sub(sub, line)
        except Exception as ex:
            print("ERROR: %s, (line(%s)"%(regex, sub))
            raise ex

        return line

    if not file_inp:
        file_inp = sys.stdin
    if not file_out:
        file_out = sys.stdout

    with io.open(file_inp) if isinstance(file_inp, str) else file_inp as fd_in:
        with io.open(file_out) if isinstance(file_out, str) else file_out as fd_out:
            line = fd_in.readline()
            while line:
                line = clean_line(line)
                print(line, file=fd_out, end='')
                line = fd_in.readline()



if __name__ == '__main__':
    clean_sphinx_only_markup(*sys.argv[1:])



#!/usr/bin/env python
#-*- coding: utf-8 -*-
#
# Copyright 2013-2019 European Commission (JRC);
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl

from __future__ import division, print_function, unicode_literals

from matplotlib import cbook, cm, pyplot as plt
from matplotlib.colors import Normalize
from numpy import ma
from numpy import polyfit, polyval
from wltp import model

import math
import numpy as np


## From http://stackoverflow.com/questions/7404116/defining-the-midpoint-of-a-colormap-in-matplotlib
#
class MidPointNorm(Normalize):
    """
    A "diverging" or "bipolar" Normalize for Colormaps on a central-point, with values going up or down.

    Example:

        norm = MidPointNorm(midpoint=3)
        imshow(X, norm=norm)
    """
    def __init__(self, midpoint=0, vmin=None, vmax=None, clip=False):
        Normalize.__init__(self, vmin, vmax, clip)
        self.midpoint = midpoint

    def __call__(self, value, clip=None):
        if clip is None:
            clip = self.clip

        result, is_scalar = self.process_value(value)

        self.autoscale_None(result)
        vmin, vmax, midpoint = self.vmin, self.vmax, self.midpoint

        if not (vmin < midpoint < vmax):
            raise ValueError("Midpoint(%s) must be between minvalue(%s) and maxvalue(%s)!"%(midpoint, vmin, vmax))
        elif vmin == vmax:
            result.fill(0) # Or should it be all masked? Or 0.5?
        elif vmin > vmax:
            raise ValueError("Maxvalue(%s) must be bigger than minvalue(%s)!"%(vmin, vmax))
        else:
            vmin = float(vmin)
            vmax = float(vmax)
            if clip:
                mask = ma.getmask(result)
                result = ma.array(np.clip(result.filled(vmax), vmin, vmax),
                                  mask=mask)

            # ma division is very slow; we can take a shortcut
            resdat = result.data

            #First scale to -1 to 1 range, than to from 0 to 1.
            resdat -= midpoint
            resdat[resdat>0] /= abs(vmax - midpoint)
            resdat[resdat<0] /= abs(vmin - midpoint)

            resdat /= 2.
            resdat += 0.5
            result = np.ma.array(resdat, mask=result.mask, copy=False)

        if is_scalar:
            result = result[0]
        return result

    def inverse(self, value):
        if not self.scaled():
            raise ValueError("Not invertible until scaled")
        vmin, vmax, midpoint = self.vmin, self.vmax, self.midpoint

        if cbook.iterable(value):
            val = ma.asarray(value)
            val = 2 * (val-0.5)
            val[val>0]  *= abs(vmax - midpoint)
            val[val<0] *= abs(vmin - midpoint)
            val += midpoint
            return val
        else:
            val = 2 * (value - 0.5)
            if val < 0:
                return  val*abs(vmin-midpoint) + midpoint
            else:
                return  val*abs(vmax-midpoint) + midpoint





def fit_straight_line(x, y):
    regress_poly = polyfit(x, y, 1)
    line_points = [x.min(), x.max()]

    return line_points, regress_poly


def plot_class_limits(axis, y):
    class_limits = model.get_class_pmr_limits()
    for limit in class_limits:
        plt.axvline(limit, color='y', linewidth=2)

    bbox = {'facecolor':'yellow', 'alpha':0.5, 'pad':4, 'linewidth':0}
    axis.text(0, y, 'class-1', style='italic', color='r',
        bbox=bbox, horizontalalignment='left', verticalalignment='top', alpha=0.8)
    axis.text(class_limits[0], y, 'class-2', style='italic', color='r',
        bbox=bbox, horizontalalignment='left', verticalalignment='bottom', alpha=0.8)
    axis.text(class_limits[1], y, 'class-3', style='italic', color='r',
        bbox=bbox, horizontalalignment='left', verticalalignment='top', alpha=0.8)


#def plot_class_parts_limits(axis, cls, y):
#    class_limits = model.get_class_parts_limits(cls)
#    for limit in class_limits:
#        plt.axvline(limit, color='y', linewidth=2)

#    ## Plot part-Limits
#    #
#    for limit in part_limits:
#        l = plt.axvline(limit, color='r', linewidth=2)
#
#        ## Add part-labels.
#        #
#        v_pos = 129.5 # trial'n error
#        if class_name == 'class1': # Acceleration scale changes!!
#            v_pos /= 2
#        bbox={'facecolor':'red', 'alpha':0.5, 'pad':4, 'linewidth':0}
#        txts = [ 'Low', 'Medium', 'High', 'Extra-high']
#        txts_pos = [0] + part_limits #[0.40, 0.67, 0.85]
#
#        for (txt, h_pos) in zip(txts, txts_pos):
#            ax1.text(h_pos + 8, v_pos, txt, style='italic',
#                bbox=bbox, size=8)


## From http://stackoverflow.com/questions/20924085/python-conversion-between-coordinates
#
def cart2pol(x, y):
    rho = np.sqrt(x**2 + y**2)
    phi = np.arctan2(y, x)
    return(rho, phi)

def cartesians_to_polarDiffs(X1, Y1, X2, Y2):
    """
    Given 2 sets of 2D-points calcs the polar euclidean-distance  and angle from 2nd-pair of points to 1st-pair.
    """
    U = X2 - X1
    V = Y2 - Y1
    DIFF, ANGLE = cart2pol(U, V)
    return U, V, DIFF, ANGLE


#############
### PLOTS ###


def plot_xy_diffs_scatter(X, Y, X_REF, Y_REF, ref_label, data_label, diff_label=None,
            data_fmt="+k", data_kws={}, ref_fmt='.g', ref_kws={},
            title=None, x_label=None, y_label=None, axes_tuple=None,
            mark_sections='classes'):
    color_diff = 'r'
    alpha = 0.8

    _, _, DIFF, _ = cartesians_to_polarDiffs(X_REF, Y_REF, X, Y)
    if axes_tuple:
        (axes, twin_axis) = axes_tuple
    else:
        axes = plt.subplot(111)

        plt.title(title)

        ## Prepare axes
        #
        axes.set_xlabel(x_label)
        axes.set_ylabel(y_label)
        axes.xaxis.grid(True)
        axes.yaxis.grid(True)

        twin_axis = axes.twinx()
        twin_axis.set_ylabel(r'$\Delta %s$' % y_label.replace('$',''), color=color_diff)
        twin_axis.tick_params(axis='y', colors=color_diff)
        twin_axis.yaxis.grid(True, color=color_diff)

        if mark_sections == 'classes':
            plot_class_limits(axes, Y.min())
        elif mark_sections == 'parts':
            raise NotImplementedError()

    ## Plot data
    #
    l_ref = axes.plot(X_REF, Y_REF, ref_fmt, label=ref_label,
        alpha=alpha, **ref_kws)
    l_data = axes.plot(X, Y, data_fmt, label=data_label, **data_kws)

    twin_axis.plot(X, DIFF, '.', color=color_diff, markersize=1.5)
    line_points, regress_poly = fit_straight_line(X, DIFF)
    l_diff = twin_axis.plot(line_points, polyval(regress_poly, line_points), '-',
        color=color_diff, label=diff_label)

    return (axes, twin_axis), (l_data, l_ref, l_diff)




def plot_xy_diffs_arrows(X, Y, X_REF, Y_REF, data_label, ref_label=None,
            data_fmt="+k", data_kws={},
            diff_label=None, diff_fmt="-r", diff_cmap=cm.hsv, diff_kws={}, #@UndefinedVariable cm.PiYG
            title=None, x_label=None, y_label=None,
            axes_tuple=None,
            mark_sections=None):
    color_diff = 'r'
    alpha = 0.9
    cm_norm = MidPointNorm()

    U, V, DIFF, ANGLE = cartesians_to_polarDiffs(X_REF, Y_REF, X, Y)

    if axes_tuple:
        (axes, twin_axis) = axes_tuple
    else:
        bottom = 0.1
        height = 0.8
        axes = plt.axes([0.1, bottom, 0.80, height])

        ## Prepare axes
        #
        axes.set_xlabel(x_label)
        axes.set_ylabel(r'$%s$' % y_label.replace('$',''))
        axes.xaxis.grid(True)
        axes.yaxis.grid(True)

        twin_axis = axes.twinx()
        twin_axis.set_ylabel(r'$\Delta %s$' % y_label.replace('$',''), color=color_diff, labelpad=0)
        twin_axis.tick_params(axis='y', colors=color_diff)
        twin_axis.yaxis.grid(True, color=color_diff)

        plt.title(title, axes=axes)
        axes_tuple = (axes, twin_axis)


    if mark_sections == 'classes':
        plot_class_limits(axes, Y.min())
    elif mark_sections == 'parts':
        raise NotImplementedError()

    ## Plot data
    #
    l_ref = axes.quiver(X, Y, U, V,
        ANGLE, cmap=diff_cmap, norm=cm_norm,
        scale_units='xy', angles='xy', scale=1,
        width=0.004, alpha=alpha,
        pivot='tip'
    )

    l_data, = axes.plot(X, Y, data_fmt, label=data_label, **data_kws)
    l_data.set_picker(3)

    l_diff = twin_axis.plot(X, V, '.', color=color_diff, markersize=0.7)
    line_points, regress_poly = fit_straight_line(X, V)
    l_diff_fitted, = twin_axis.plot(line_points, polyval(regress_poly, line_points), diff_fmt,
        label=diff_label, **diff_kws)

    return axes_tuple, (l_data, l_ref, l_diff, l_diff_fitted)

## http://stackoverflow.com/questions/13306519/get-data-from-plot-with-matplotlib
class DataCursor(object):
    """
    Use::

        fig.canvas.mpl_connect('pick_event', DataCursor(plt.gca()))
    """
    x, y = 0.0, 0.0
    xoffset, yoffset = -20, 20

    def __init__(self, ax=None,
                 text_template='x: {x:0.2f}\ny: {y:0.2f}\n{txt}',
                 annotations=None):
        self.ax = ax or plt.gca()
        self.text_template = text_template
        self.annotations = annotations or []
        self.annotation = ax.annotate(self.text_template,
                xy=(self.x, self.y), xytext=(self.xoffset, self.yoffset),
                textcoords='offset points', ha='right', va='bottom',
                bbox=dict(boxstyle='round,pad=0.5', fc='yellow', alpha=0.5),
                arrowprops=dict(arrowstyle='->', connectionstyle='arc3,rad=0')
                )
        self.annotation.set_visible(False)

    def __call__(self, event):
        if event.mouseevent.xdata is not None:
            try:
                ind = event.ind[0] if isinstance(event.ind, np.ndarray) else event.ind
                x = event.artist.get_xdata()[ind]
                y = event.artist.get_ydata()[ind]
                try:
                    annotation = self.annotations[ind]
                except:
                    annotation = ''
                self.annotation.xy = (event.mouseevent.xdata, event.mouseevent.ydata)
                self.annotation.set_text(self.text_template.format(x=x, y=y, txt=annotation))
                self.annotation.set_visible(True)
                event.canvas.draw()
            except:
                pass



#!/usr/bin/env python
#-*- coding: utf-8 -*-
#
# Copyright 2013-2019 European Commission (JRC);
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl

import os

from matplotlib import pyplot as plt
from wltp import plots
from wltp.test import wltp_db_tests as wltpdb


def prepare_data(gened_column, heinz_column):
    gened_column='rpm' 
    heinz_column='n'
    vehdata = wltpdb.aggregate_single_columns_means(gened_column, heinz_column)
    vehdata['pmr'] = 1000.0 * vehdata['rated_power'] / vehdata['kerb_mass']

    return vehdata.pmr, vehdata.gened, vehdata.heinz


if __name__ == '__main__':
    os.chdir(os.path.join(wltpdb.mydir, wltpdb.samples_dir))
    
    (X, Y, Y_REF) = prepare_data(gened_column='rpm', heinz_column='n')

    
    kws = dict(data_fmt='ok', data_kws=dict(markersize=4.5, fillstyle='none'), )
    plots.plot_xy_diffs_arrows(
        X, Y, X, Y_REF, 
        title="Python(arrow-heads) vs AccessDb-2secAccel (arrow-tails)",
        x_label=r'$PMR [W/kg]$',
        y_label='EngineSpeed [rpm]', 
        ref_label='Access-db', data_label='Python',
        mark_sections='classes',
        **kws
    )
    plt.show()
    



#!/usr/bin/env python
#-*- coding: utf-8 -*-
#
# Copyright 2013-2019 European Commission (JRC);
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl
"""A :dfn:`pandas-model` is a tree of strings, numbers, sequences, dicts, pandas instances and resolvable
URI-references, implemented by :class:`Pandel`. """

from __future__ import division, unicode_literals

import abc
from collections.abc import Mapping, Sequence
from collections import OrderedDict, namedtuple, defaultdict
import contextlib
import numbers
import re

from jsonschema import Draft3Validator, Draft4Validator, ValidationError
import jsonschema
from jsonschema.exceptions import SchemaError
from pandas.core.generic import NDFrame
from six import string_types
try:
    from unittest.mock import MagicMock
except ImportError:
    from mock import MagicMock

import numpy as np
import pandas as pd


try:
    from urllib.parse import urljoin
except ImportError:
    from urlparse import urljoin


_value_with_units_regex = re.compile(r'''^\s*
                                        (?P<name>[^([]+?)   # column-name
                                        \s*
                                        (?P<units>          # start parenthesized-units optional-group
                                            \[              # units enclosed in []
                                                [^\]]*
                                            \]
                                            |
                                            \(              # units enclosed in ()
                                                [^)]*
                                            \)
                                        )?                  # end parenthesized-units
                                        \s*$''', re.X)
_units_cleaner_regex = re.compile(r'^[([]]|[\])]$')

def parse_value_with_units(arg):
    """
    Something like `value [units]` or `foo (bar/krow).

    :return: dict(name, units) or None if unparseable as column-suntax
    """

    m = _value_with_units_regex.match(arg)
    res = m.groupdict()
    units = res['units']
    if units:
        res['units'] = _units_cleaner_regex.sub('', units)
    return res



class ModelOperations(namedtuple('ModelOperations', 'inp out conv')):
    """
    Customization functions for traversing, I/O, and converting self-or-descendant branch (sub)model values.
    """
    def __new__(cls, inp=None, out=None, conv=None):
        """

        :param list inp:    the `args-list` to :meth:`Pandel._read_branch()`

        :param out:         The args to :meth:`Pandel._write_branch()`, that may be specified either as:

                            * an `args-list`, that will apply for all model data-types (lists, dicts & pandas),
                            * a map of ``type`` --> ``args-list``, where the ``None`` key is the *catch-all* case,
                            * a function returning the `args-list` for some branch-value,
                              with signature: ``def get_write_branch_args(branch)``.

        :param conv:        The conversion-functions (:dfn:`convertors`) for the various model's data-types.
                            The convertors have signature ``def convert(branch)``, and they may be
                            specified either as:

                            * a map of ``(from_type, to_type)`` --> ``conversion_func()``, where the ``None`` key
                              is the *catch-all* case,
                            * a "master-switch" function returning the appropriate convertor
                              depending on the requested conversion.
                              The master-function's signature is ``def get_convertor(from_branch, to_branch)``.

                            The minimum convertors demanded by :class:`Pandel` are (at least, check the code for more):

                            * DataFrame  <--> dict
                            * Series     <--> dict
                            * ndarray    <--> list
        """

        return super(ModelOperations, cls).__new__(cls, inp, out, conv)

    def choose_out_args(self, branch):
        pass

    def choose_convertor(self, from_type, to_type):
        pass

class PathMaps(object):
    """
    Cascade prefix-mapping of json-paths to any values (here :class:`ModelOperations`.
    """
    pass



ValidatorBase = jsonschema.validators.create({}) # Workaround https://github.com/Julian/jsonschema/issues/178
class PandelVisitor(ValidatorBase):
    """
    A customized :class:`Draft4Validator` suporting instance-trees with pandas and numpy objects, natively.

    Any pandas or numpy instance (for example ``obj``) is treated like that:

    +----------------------------+-----------------------------------------+
    |        Python Type         |     JSON Equivalence                    |
    +============================+=========================================+
    | :class:`pandas.DataFrame`  | as ``object`` *json-type*, with         |
    |                            | ``obj.columns`` as *keys*, and          |
    |                            | ``obj[col].values`` as *values*         |
    +----------------------------+-----------------------------------------+
    | :class:`pandas.Series`     | as ``object`` *json-type*, with         |
    |                            | ``obj.index`` as *keys*, and            |
    |                            | ``obj.values`` as *values*              |
    +----------------------------+-----------------------------------------+
    | :class:`np.ndarray`,       | as ``array`` *json-type*                |
    | :class:`list`,             |                                         |
    | :class:`tuple`             |                                         |
    +----------------------------+-----------------------------------------+

    Note that the value of each dataFrame column is a :``ndarray`` instances.

    The simplest validations of an object or a pandas-instance is like this:

        >>> import pandas as pd

        >>> schema = {
        ...     'type': 'object',
        ... }
        >>> pv = PandelVisitor(schema)

        >>> pv.validate({'foo': 'bar'})
        >>> pv.validate(pd.Series({'foo': 1}))
        >>> pv.validate([1,2])                                       ## A sequence is invalid here.
        Traceback (most recent call last):
        ...
        jsonschema.exceptions.ValidationError: [1, 2] is not of type 'object'
        <BLANKLINE>
        Failed validating 'type' in schema:
            {'type': 'object'}
        <BLANKLINE>
        On instance:
            [1, 2]


    Or demanding specific properties with ``required`` and no ``additionalProperties``:

        >>> schema = {
        ...     'type':     'object',
        ...     'required': ['foo'],
        ...    'additionalProperties': False,
        ...    'properties': {
        ...        'foo': {}
        ...    }
        ... }
        >>> pv = PandelVisitor(schema)

        >>> pv.validate(pd.Series({'foo': 1}))
        >>> pv.validate(pd.Series({'foo': 1, 'bar': 2}))             ## Additional 'bar' is present!
        Traceback (most recent call last):
        ...
        jsonschema.exceptions.ValidationError: Additional properties are not allowed ('bar' was unexpected)
        <BLANKLINE>
        Failed validating 'additionalProperties' in schema:
            {'additionalProperties': False,
             'properties': {'foo': {}},
             'required': ['foo'],
             'type': 'object'}
        <BLANKLINE>
        On instance:
            bar    2
            foo    1
            dtype: int64

        >>> pv.validate(pd.Series({}))                               ## Required 'foo' missing!
        Traceback (most recent call last):
        ...
        jsonschema.exceptions.ValidationError: 'foo' is a required property
        <BLANKLINE>
        Failed validating 'required' in schema:
            {'additionalProperties': False,
             'properties': {'foo': {}},
             'required': ['foo'],
             'type': 'object'}
        <BLANKLINE>
        On instance:
            Series([], dtype: float64)

    """
    def __init__(self, schema, types=(), resolver=None, format_checker=None, skip_meta_validation=False):
        super(PandelVisitor, self).__init__(schema, types, resolver, format_checker)

        self._types.update({
            "number":   (numbers.Number, np.number), ## type(np.nan) == builtins.float! FIXME, are numpy-numbers --> json-types OK??
            "integer":  (int, np.integer),
            "boolean":  (bool, np.bool_), #, np.bool8),
            "array":    (list, tuple, np.ndarray),
            "object" :  (dict, pd.DataFrame, pd.Series)
        })

        ## Setup Draft4/3 validation
        #
        # Meta-validate schema
        #    with original validators (and not self)
        #    because this class inherits an empty (schema/rules) validator.
        validator_class = jsonschema.validators.validator_for(schema)  ## Falls back to 'Draft4' if no `$schema` exists.
        self.VALIDATORS = validator_class.VALIDATORS.copy()
        self.META_SCHEMA = validator_class.META_SCHEMA
        self.VALIDATORS.update({
            'items':                PandelVisitor._rule_items,
            'additionalProperties': PandelVisitor._rule_additionalProperties,
            'additionalItems':      PandelVisitor._rule_additionalItems,
        })
        if validator_class == Draft3Validator:
            self.VALIDATORS.update({
                'properties':           PandelVisitor._rule_properties_draft3,
            })
        else:
            self.VALIDATORS.update({
                'properties':           PandelVisitor._rule_properties_draft4,
                'required':             PandelVisitor._rule_required_draft4,
            })

        self.old_scopes = []

        ## Cannot use ``validator_class.check_schema()`` because
        #    need to relay my args to ``validator_class.__init__()``.
        # Even better use myself, that i'm fatser (kind of...).
        if not skip_meta_validation:
            for error in self.iter_errors(schema, validator_class.META_SCHEMA):
                raise SchemaError.create_from(error)

    ##################################
    ############ Visiting ###########
    ##################################

    def _get_iprop(self, instance, prop):
        val = instance[prop]
        if isinstance(val, NDFrame):
            val = val.values
        return val

    def _is_iprop_in(self, instance, prop):
        return prop in instance.keys()

    def _iter_iprop_names(self, instance):
        return instance.keys()

    def _iter_iprop_pairs(self, instance):
        if isinstance(instance, pd.DataFrame):
            return ((k, v.values) for k, v in instance.iteritems())
        if isinstance(instance, pd.Series):
            return instance.iteritems()
        return instance.items()

    def _iter_iitems(self, instance):
        return instance



    def iter_errors(self, instance, _schema=None):
        if _schema is None:
            _schema = self.schema

        scope = _schema.get("id")
        if scope:
            self.resolver.push_scope(scope)

        ref = _schema.get("$ref")
        if ref is not None:
            validators = [("$ref", ref)]
        else:
            validators = self._iter_iprop_pairs(_schema)

        for k, v in validators:
            validator = self.VALIDATORS.get(k)
            if validator is None:
                continue

            errors = validator(self, v, instance, _schema) or ()
            for error in errors:
                # set details if not already set by the called fn
                error._set(
                    validator=k,
                    validator_value=v,
                    instance=instance,
                    schema=_schema,
                )
                if k != "$ref":
                    error.schema_path.appendleft(k)
                yield error

        if scope:
            self.resolver.pop_scope()


    ##################################
    ############# Rules ##############
    ##################################

    def _rule_properties_draft4(self, sprops, instance, schema):
        if not self.is_type(instance, "object"):
            return

        iprops = set(self._iter_iprop_names(instance))
        for prop in iprops & set(sprops.keys()):
            subschema = sprops[prop]
            for error in self.descend(
                self._get_iprop(instance, prop),
                subschema,
                path=prop,
                schema_path=prop,
            ):
                yield error

    def _rule_properties_draft3(self, properties, instance, schema):
        if not self.is_type(instance, "object"):
            return

        for prop, subschema in self._iter_iprop_pairs(properties):
            if self._is_iprop_in(instance, prop):
                for error in self.descend(
                    self._get_iprop(instance, prop),
                    subschema,
                    path=prop,
                    schema_path=prop,
                ):
                    yield error
            elif subschema.get("required", False):
                error = ValidationError("%r is a required prop" % prop)
                error._set(
                    validator="required",
                    validator_value=subschema["required"],
                    instance=instance,
                    schema=schema,
                )
                error.path.appendleft(prop)
                error.schema_path.extend([prop, "required"])
                yield error

    def _rule_items(self, items, instance, schema):
        if not self.is_type(instance, "array"):
            return

        if self.is_type(items, "object"):
            for index, item in enumerate(self._iter_iitems(instance)):
                for error in self.descend(item, items, path=index):
                    yield error
        else:
            for (index, item), subschema in zip(enumerate(self._iter_iitems(instance)), items):
                for error in self.descend(
                    item, subschema, path=index, schema_path=index,
                ):
                    yield error

    def _rule_additionalProperties(self, aP, instance, schema):
        if not self.is_type(instance, 'object'):
            return

        sprops = schema.get("properties", {})
        patterns = "|".join(schema.get("patternProperties", {}))
        extras = set()
        for iprop in self._iter_iprop_names(instance):
            if iprop not in sprops and \
                    not patterns or not re.search(patterns, iprop):
                extras.add(iprop)

        if extras:
            if self.is_type(aP, "object"):
                for extra in extras:
                    for error in self.descend(self._get_iprop(instance, extra), aP, path=extra):
                        yield error
            elif not aP:
                yield ValidationError(
                    "Additional properties are not allowed (%s %s unexpected)" %
                    jsonschema._utils.extras_msg(extras))

    def _rule_additionalItems(self, aI, instance, schema):
        if (
            not self.is_type(instance, "array") or
            self.is_type(schema.get("items", {}), "object")
        ):
            return

        len_items = len(schema.get("items", []))
        if self.is_type(aI, "object"):
            for index, item in enumerate(instance[len_items:], start=len_items):
                for error in self.descend(item, aI, path=index):
                    yield error
        elif not aI and len(instance) > len_items:
            yield ValidationError(
                "Additional items are not allowed (%s %s unexpected)" %
                jsonschema._utils.extras_msg(instance[len(schema.get("items", [])):])
            )



    def _rule_required_draft4(self, required, instance, schema):
        if self.is_type(instance, 'object'):
            for sprop in required:
                if not self._is_iprop_in(instance, sprop):
                    yield ValidationError("%r is a required property" % sprop)



class Pandel(object):
    """
    Builds, validates and stores a *pandas-model*, a mergeable stack of JSON-schema abiding trees of
    strings and numbers, assembled with

    * sequences,
    * dictionaries,
    * :class:`pandas.DataFrame`,
    * :class:`pandas.Series`, and
    * URI-references to other model-trees.



    .. _pandel-overview:

    **Overview**

    The **making of a model** involves, among others, schema-validating, reading :dfn:`subtree-branches`
    from URIs, cloning, converting and merging multiple :dfn:`sub-models` in a single :dfn:`unified-model` tree,
    without side-effecting given input.
    All these happen in 4+1 steps::

                       ....................... Model Construction .................
          ------------ :  _______    ___________                                  :
         / top_model /==>|Resolve|->|PreValidate|-+                               :
         -----------'  : |___0___|  |_____1_____| |                               :
          ------------ :  _______    ___________  |   _____    ________    ______ :   --------
         / base-model/==>|Resolve|->|PreValidate|-+->|Merge|->|Validate|->|Curate|==>/ model /
         -----------'  : |___0___|  |_____1_____|    |_ 2__|  |___3____|  |__4+__|:  -------'
                       ............................................................

    All steps are executed "lazily" using generators (with :keyword:`yield`).
    Before proceeding to the next step, the previous one must have completed successfully.
    That way, any ad-hoc code in building-step-5(*curation*), for instance, will not suffer a horrible death
    due to badly-formed data.

    [TODO] The **storing of a model** simply involves distributing model parts into different files and/or formats,
    again without side-effecting the unified-model.



    .. _pandel-building-model:

    **Building model**

    Here is a detailed description of each building-step:

    1.  :meth:`_resolve` and substitute any `json-references <http://tools.ietf.org/html/draft-pbryan-zyp-json-ref-03>`_
        present in the submodels with content-fragments fetched from the referred URIs.
        The submodels are **cloned** first, to avoid side-effecting them.

        Although by default a combination of *JSON* and *CSV* files is expected, this can be customized,
        either by the content in the json-ref, within the model (see below), or
        as :ref:`explained  <pandel-customization>` below.

        The **extended json-refs syntax** supported provides for passing arguments into :meth:`_read_branch()`
        and :meth:`_write_branch()` methods.  The syntax is easier to explain by showing what
        the default :attr:`_global_cntxt` corresponds to, for a ``DataFrame``::

            {
              "$ref": "http://example.com/example.json#/foo/bar",
              "$inp": ["AUTO"],
              "$out": ["CSV", "encoding=UTF-8"]
            }

        And here what is required to read and (later) store into a HDF5 local file with a predefined name::

            {
              "$ref": "file://./filename.hdf5",
              "$inp": ["AUTO"],
              "$out": ["HDF5"]
            }

        .. Warning:: Step NOT IMPLEMENTED YET!


    2.  Loosely :meth:`_prevalidate` each sub-model separately with :term:`json-schema`,
        where any pandas-instances (DataFrames and Series) are left as is.
        It is the duty of the developer to ensure that the prevalidation-schema is *loose enough* that
        it allows for various submodel-forms, prior to merging, to pass.


    3.  Recursively **clone**  and :meth:`_merge` sub-models in a single unified-model tree.
        Branches from sub-models higher in the stack override the respective ones from the sub-models below,
        recursively.  Different object types need to be **converted** appropriately (ie. merging
        a ``dict`` with a ``DataFrame`` results into a ``DataFrame``, so the dictionary has to convert
        to dataframe).

        The required **conversions** into pandas classes can be customized as :ref:`explained  <pandel-customization>`
        below.  Series and DataFrames cannot merge together, and Sequences do not merge
        with any other object-type (themselfs included), they just "overwrite".

        The default convertor-functions defined both for submodels and models are listed in the following table:

        ============ ========== =========================================
            From:       To:                  Method:
        ============ ========== =========================================
         dict        DataFrame  ``pd.DataFrame``  (the constructor)
         DataFrame   dict       ``lambda df: df.to_dict('list')``
         dict        Series     ``pd.Series``     (the constructor)
         Series      dict       :meth:`lambda sr: sr.to_dict()`
        ============ ========== =========================================


    4.  Strictly json-:meth:`_validate` the unified-model (ie enforcing ``required`` schema-rules).

        The required **conversions** from pandas classes can be customized as :ref:`explained  <pandel-customization>`
        below.

        The default convertor-functions are the same as above.


    5.  (Optionally) Apply the :meth:`_curate` functions on the the model to enforce dependencies and/or any
        ad-hoc generation-rules among the data.  You can think of bash-like expansion patterns,
        like ``${/some/path:=$HOME}`` or expressions like ``%len(../other/path)``.



    .. _pandel-storing:

    **Storing model**

    When storing model-parts, if unspecified, the filenames to write into will be deduced from the jsonpointer-path
    of the ``$out``'s parent, by substituting "strange" chars with undescores(``_``).

    .. Warning:: Functionality NOT IMPLEMENTED YET!



    .. _pandel-customization:

    **Customization**

    Some operations within steps (namely *conversion* and *IO*) can be customized by the following means
    (from lower to higher precedance):

    a.  The global-default :class:`ModelOperations` instance on the :attr:`_global_cntxt`,
        applied on both submodels and unified-model.

        For example to channel the whole reading/writing of models through
        `HDF5 <http://pandas.pydata.org/pandas-docs/stable/io.html#io-hdf5>`_ data-format, it would suffice
        to modify the :attr:`_global_cntxt` like that::

            pm = FooPandelModel()                        ## some concrete model-maker
            io_args = ["HDF5"]
            pm.mod_global_operations(inp=io_args, out=io_args)

    b.  [TODO] Extra-properties on the json-schema applied on both submodels and unified-model for the specific path defined.
        The supported properties are the non-functional properties of :class:`ModelOperations`.

    d.  Specific-properties regarding *IO* operations within each submodel - see the *resolve* building-step,
        above.

    c.  Context-maps of ``json_paths`` --> :class:`ModelOperations` instances, installed by :meth:`add_submodel()` and
        :attr:`unified_contexts` on the model-maker.  They apply to self-or-descedant subtree of each model.

        The `json_path` is a strings obeying a simplified :term:`json-pointer` syntax (no char-normalizations yet),
        ie ``/some/foo/1/pointer``.  An empty-string(``''``) matches all model.

        When multiple convertors match for a model-value, the selected convertor to be used is the most specific one
        (the one with longest prefix).  For instance, on the model::

            [ { "foo": { "bar": 0 } } ]


        all of the following would match the ``0`` value:

        - the global-default :attr:`_global_cntxt`,
        - ``/``, and
        - ``/0/foo``

        but only the last's context-props will be applied.



    .. _Attributes:

    **Atributes**

    .. Attribute:: model

        The model-tree that will receive the merged submodels after :meth:`build()` has been invoked.
        Depending on the submodels, the top-value can be any of the supported model data-types.


    .. Attribute:: _submodel_tuples

        The stack of (``submodel``, ``path_ops``) tuples. The list's 1st element is the :dfn:`base-model`,
        the last one, the :dfn:`top-model`.  Use the :meth:`add_submodel()` to build this list.


    .. Attribute:: _global_cntxt

        A :class:`ModelOperations` instance acting as the global-default context for the unified-model and all submodels.
        Use :meth:`mod_global_operations()` to modify it.


    .. Attribute:: _curate_funcs

        The sequence of *curate* functions to be executed as the final step by :meth:`_curate()`.
        They are "normal" functions (not generators) with signature::

            def curate_func(model_maker):
                pass      ## ie: modify ``model_maker.model``.

        Better specify this list of functions on construction time.


    .. Attribute:: _errored

            An internal boolean flag that becomes ``True`` if any build-step has failed,
            to halt proceeding to the next one.  It is ``None`` if build has not started yet.


    .. _pandel-examples:

    **Examples**

    The basic usage requires to subclass your own model-maker, just so that a *json-schema*
    is provided for both validation-steps, 2 & 4:

        >>> from collections import OrderedDict as od                           ## Json is better with stable keys-order

        >>> class MyModel(Pandel):
        ...     def _get_json_schema(self, is_prevalidation):
        ...         return {                                                    ## Define the json-schema.
        ...             '$schema': 'http://json-schema.org/draft-04/schema#',
        ...             'required': [] if is_prevalidation else ['a', 'b'],     ## Prevalidation is more loose.
        ...             'properties': {
        ...                 'a': {'type': 'string'},
        ...                 'b': {'type': 'number'},
        ...                 'c': {'type': 'number'},
        ...             }
        ...         }


    Then you can instanciate it and add your submodels:

        >>> mm = MyModel()
        >>> mm.add_submodel(od(a='foo', b=1))                                   ## submodel-1 (base)
        >>> mm.add_submodel(pd.Series(od(a='bar', c=2)))                        ## submodel-2 (top-model)


    You then have to build the final unified-model (any validation errors would be reported at this point):

        >>> mdl = mm.build()

    Note that you can also access the unified-model in the :attr:`model` attribute.
    You can now interogate it:

        >>> mdl['a'] == 'bar'                       ## Value overridden by top-model
        True
        >>> mdl['b'] == 1                           ## Value left intact from base-model
        True
        >>> mdl['c'] == 2                           ## New value from top-model
        True


    Lets try to build with invalid submodels:

        >>> mm = MyModel()
        >>> mm.add_submodel({'a': 1})               ## According to the schema, this should have been a string,
        >>> mm.add_submodel({'b': 'string'})        ## and this one, a number.

        >>> sorted(mm.build_iter(), key=lambda ex: ex.message)                   ## Fetch a list with all validation errors.
        [<ValidationError: "'string' is not of type 'number'">,
         <ValidationError: "1 is not of type 'string'">,
         <ValidationError: 'Gave-up building model after step 1.prevalidate (out of 4).'>]

        >>> mdl = mm.model
        >>> mdl is None                                     ## No model constructed, failed before merging.
        True


    And lets try to build with valid submodels but invalid merged-one:

        >>> mm = MyModel()
        >>> mm.add_submodel({'a': 'a str'})
        >>> mm.add_submodel({'c': 1})

        >>> sorted(mm.build_iter(), key=lambda ex: ex.message)        ## Missing required('b') prop rom merged-model.
        [<ValidationError: "'b' is a required property">,
         <ValidationError: 'Gave-up building model after step 3.validate (out of 4).'>]

    """

    __metaclass__ = abc.ABCMeta

    def __init__(self, curate_funcs=()):
        """

        :param sequence curate_funcs:   See :attr:`_curate_funcs`.
        """

        self.model          = None
        self._errored       = None
        self._submodel_tuples   = []
        self._curate_funcs  = curate_funcs
        self._global_cntxt  = []
        self._unified_contexts = None


    def mod_global_operations(self, operations=None, **cntxt_kwargs):
        """

        Since it is the fall-back operation for *conversions* and *IO* operation, it must exist and have
        all its props well-defined for the class to work correctly.

        :param ModelOperations operations:  Replaces values of the installed context with
                                            non-empty values from this one.
        :param cntxt_kwargs:                Replaces the keyworded-values on the existing `operations`.
                                            See :class:`ModelOperations` for supported keywords.
        """
        if operations:
            assert isinstance(operations, ModelOperations), (type(operations), operations)
            self._global_cntxt = operations
        self._global_cntxt._replace(**cntxt_kwargs)


    @property
    def unified_contexts(self):
        """
        A map of ``json_paths`` --> :class:`ModelOperations` instances acting on the unified-model.
        """
        return self._unified_contexts
    @unified_contexts.setter
    def unified_contexts(self, path_ops):
        assert isinstance(path_ops, Mapping), (type(path_ops), path_ops)
        self._unified_contexts = path_ops

    def _select_context(self, path, branch):
        """
        Finds which context to use while visiting model-nodes, by enforcing the precedance-rules described
        in the :ref:`Customizations  <pandel-customization>`.

        :param str path:    the branch's jsonpointer-path
        :param str branch:  the actual branch's node
        :return:            the selected :class:`ModelOperations`
        """
        pass


    def _read_branch(self):
        """
        Reads model-branches during *resolve* step.
        """
        pass # TODO: impl read_branch()
    def _write_branch(self):
        """
        Writes model-branches during *distribute* step.
        """
        pass # TODO: impl write_branch()

    def _get_json_schema(self, is_prevalidation):
        """
        :return: a json schema, more loose when `prevalidation` for each case
        :rtype: dictionary
        """
        ## TODO: Make it a factory o
        pass

    def _rule_AdditionalProperties(self, validator, aP, required, instance, schema):
        properties = schema.get("properties", {})
        patterns = "|".join(schema.get("patternProperties", {}))
        extras = set()
        for prop in instance:
            if prop not in properties:
                if patterns and re.search(patterns, prop):
                    continue
                extras.add(prop)

        if validator.is_type(aP, "object"):
            for extra in extras:
                for error in validator.descend(instance[extra], aP, path=extra):
                    yield error
        elif not aP and extras:
            error = "Additional properties are not allowed (%s %s unexpected)"
            yield ValidationError(error % jsonschema._utils.extras_msg(extras))


    def _rule_Required(self, validator, required, instance, schema):
        if (validator.is_type(instance, "object") or
                validator.is_type(instance, "DataFrame") or
                 validator.is_type(instance, "Series")):
            for prop in required:
                if prop not in instance:
                    yield ValidationError("%r is a required property" % prop)


    def _get_model_validator(self, schema):

        validator = Draft4Validator(schema)
        validator._types.update({"ndarray": np.ndarray, "DataFrame" : pd.DataFrame, 'Series':pd.Series})
        validator.VALIDATORS['DataFrame'] = self._rule_Required

        return validator


    def _validate_json_model(self, schema, mdl):
        validator = self._get_model_validator(schema)
        for err in validator.iter_errors(mdl):
            self._errored = True
            yield err

    def _clone_and_merge_submodels(self, a, b, path=''):
        """' Recursively merge b into a, cloning both. """

        if isinstance(a, pd.DataFrame) or isinstance(b, pd.DataFrame):
            a = pd.DataFrame() if a is None else pd.DataFrame(a)
            b = pd.DataFrame() if b is None else pd.DataFrame(b)

            a.update(b) #, 'outer') NOT IMPL YET
            extra_b_items = list(set(b.columns) - set(a.columns))
            a[extra_b_items] = b[extra_b_items]

        elif isinstance(a, pd.Series) or isinstance(b, pd.Series):
            a = pd.Series() if a is None else pd.Series(a)
            b = pd.Series() if b is None else pd.Series(b)
            #a.update(b) # DOES NOT append extra keys!
            a = b.combine_first(a)

        elif isinstance(a, Mapping) or isinstance(b, Mapping):
            a = OrderedDict() if a is None else OrderedDict(a)
            b = OrderedDict() if b is None else OrderedDict(b)

            for key in b:
                b_val = b[key]
                if key in a:
                    val = self._clone_and_merge_submodels(a[key], b_val, '%s/%s'%(path, key))
                else:
                    val = b_val
                a[key] = val

        elif (isinstance(a, Sequence) and not isinstance(a, string_types)) or \
                (isinstance(b, Sequence) and not isinstance(b, string_types)):
            if not b is None:
                val = b
            else:
                val = a

            l = list()
            for (i, item) in enumerate(val):
                l.append(self._clone_and_merge_submodels(item, None, '%s[%i]'%(path, i)))
            a = l

        elif a is None and b is None:
            raise ValidationError("Cannot merge Nones at path(%s)!" % path)

        else:
            if not b is None:
                a = b

        return a

    def _resolve(self):
        "Step-1"
        if False:
            yield

    def _prevalidate(self):
        "Step-1"
        for (mdl, path_ops) in self._submodel_tuples:
            schema = self._get_json_schema(is_prevalidation=True)
            for err in self._validate_json_model(schema, mdl):
                yield err

    def _merge(self):
        "Step-2"
        for (mdl, path_ops) in self._submodel_tuples:
            self.model = self._clone_and_merge_submodels(self.model, mdl)
        if False:
            yield       ## Just mark method as generator.

    def _validate(self):
        "Step-3"
        schema = self._get_json_schema(is_prevalidation=False)
        for err in self._validate_json_model(schema, self.model):
            yield err

    def _curate(self):
        "Step-4:  Invokes any curate-functions found in :attr:`_curate_funcs`."
        if False:
            yield       ## To be overriden, just mark method as generator.
        for curfunc in self._curate_funcs:
            curfunc(self)

    def add_submodel(self, model, path_ops=None):
        """
        Pushes on top a submodel, along with its context-map.

        :param model:               the model-tree (sequence, mapping, pandas-types)
        :param dict path_ops:       A map of ``json_paths`` --> :class:`ModelOperations` instances acting on the
                                    unified-model.  The `path_ops` may often be empty.

        **Examples**

        To change the default DataFrame --> dictionary convertor for a submodel, use the following:

            >>> mdl = {'foo': 'bar'}
            >>> submdl = ModelOperations(mdl, conv={(pd.DataFrame, dict): lambda df: df.to_dict('record')})

        """

        if path_ops:
            assert isinstance(path_ops, Mapping), (type(path_ops), path_ops)

        return self._submodel_tuples.append((model, path_ops))


    def build_iter(self):
        """
        Iteratively build model, yielding any problems as :class:`ValidationError` instances.

        For debugging, the unified model at :attr:`model` my contain intermediate results at any time,
        even if construction has failed.  Check the :attr:`_errored` flag if neccessary.
        """

        steps = [
            (self._prevalidate, 'prevalidate'),
            (self._merge,       'merge'),
            (self._validate,    'validate'),
            (self._curate,      'curate'),
        ]
        self._errored = False
        self.model = None

        for (i, (step, step_name)) in enumerate(steps, start=1):
            try:
                for err in step():
                    yield err
            except ValidationError as ex:
                self._errored = True
                yield ex

            except Exception as ex:
                self._errored = True

                nex = ValidationError('Model step-%i(%s) failed due to: %s'%(i, step_name, ex))
                nex.cause = ex

                yield nex

            if self._errored:
                yield ValidationError('Gave-up building model after step %i.%s (out of %i).'%(i, step_name, len(steps)))
                break

    def build(self):
        """
        Attempts to build the model by exhausting :meth:`build_iter()`, or raises its 1st error.

        Use this method when you do not want to waste time getting the full list of errors.
        """

        err = next(self.build_iter(), None)
        if err:
            raise err

        return self.model


class JsonPointerException(Exception):
    pass

def jsonpointer_parts(jsonpointer):
    """
    Iterates over the ``jsonpointer`` parts.

    :param str jsonpointer: a jsonpointer to resolve within document
    :return: a generator over the parts of the json-pointer

    :author: Julian Berman, ankostis
    """

    if jsonpointer:
        parts = jsonpointer.split(u"/")
        if parts.pop(0) != '':
            raise JsonPointerException('Location must starts with /')

        for part in parts:
            part = part.replace(u"~1", u"/").replace(u"~0", u"~")

            yield part

_scream = object()
def resolve_jsonpointer(doc, jsonpointer, default=_scream):
    """
    Resolve a ``jsonpointer`` within the referenced ``doc``.

    :param doc: the referrant document
    :param str jsonpointer: a jsonpointer to resolve within document
    :return: the resolved doc-item or raises :class:`RefResolutionError`

    :author: Julian Berman, ankostis
    """
    for part in jsonpointer_parts(jsonpointer):
        if isinstance(doc, Sequence):
            # Array indexes should be turned into integers
            try:
                part = int(part)
            except ValueError:
                pass
        try:
            doc = doc[part]
        except (TypeError, LookupError):
            if default is _scream:
                raise JsonPointerException("Unresolvable JSON pointer(%r)@(%s)" % (jsonpointer, part))
            else:
                return default

    return doc


def set_jsonpointer(doc, jsonpointer, value, object_factory=dict):
    """
    Resolve a ``jsonpointer`` within the referenced ``doc``.

    :param doc: the referrant document
    :param str jsonpointer: a jsonpointer to the node to modify
    :raises: JsonPointerException (if jsonpointer empty, missing, invalid-contet)
    """


    parts = list(jsonpointer_parts(jsonpointer))

    ## Will scream if used on 1st iteration.
    #
    pdoc = None
    ppart = None
    for i, part in enumerate(parts):
        if isinstance(doc, Sequence) and not isinstance(doc, str):
            ## Array indexes should be turned into integers
            #
            doclen = len(doc)
            if part == '-':
                part = doclen
            else:
                try:
                    part = int(part)
                except ValueError:
                    raise JsonPointerException("Expected numeric index(%s) for sequence at (%r)[%i]" % (part, jsonpointer, i))
                else:
                    if part > doclen:
                        raise JsonPointerException("Index(%s) out of bounds(%i) of (%r)[%i]" % (part, doclen, jsonpointer, i))
        try:
            ndoc = doc[part]
        except (LookupError):
            break  ## Branch-extension needed.
        except (TypeError): # Maybe indexing a string...
            ndoc = object_factory()
            pdoc[ppart] = ndoc
            doc = ndoc
            break  ## Branch-extension needed.

        doc, pdoc, ppart = ndoc, doc, part
    else:
        doc = pdoc # If loop exhausted, cancel last assignment.

    ## Build branch with value-leaf.
    #
    nbranch = value
    for part2 in reversed(parts[i+1:]):
        ndoc = object_factory()
        ndoc[part2] = nbranch
        nbranch = ndoc

    ## Attach new-branch.
    try:
        doc[part] = nbranch
    except IndexError: # Inserting last sequence-element raises IndexError("list assignment index out of range")
        doc.append(nbranch)

#    except (IndexError, TypeError) as ex:
#        #if isinstance(ex, IndexError) or 'list indices must be integers' in str(ex):
#        raise JsonPointerException("Incompatible content of JSON pointer(%r)@(%s)" % (jsonpointer, part))
#        else:
#            doc = {}
#            parent_doc[parent_part] = doc
#            doc[part] = value


def build_all_jsonpaths(schema):
    ## Totally quick an dirty, TODO: Use json-validator to build all json-paths.
    forks = ['oneOf', 'anyOf', 'allOf']
    def _visit(schema, path, paths):
        for f in forks:
            objlist = schema.get(f)
            if objlist:
                for obj in schema.get(f):
                    _visit(obj, path, paths)

        props = schema.get('properties')
        if props:
            for p, obj in props.items():
                _visit(obj, path + '/' +p, paths)
        else:
            paths.append(path)

    paths = []
    _visit(schema, '', paths)

    return paths

if __name__ == '__main__':
    raise "Not runnable!"



#!/usr/bin/env python
#-*- coding: utf-8 -*-
#
# Copyright 2013-2019 European Commission (JRC);
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl

from __future__ import division, unicode_literals

import argparse
import sys, os
import unittest


##############
#  Compatibility
#
try:
    assertRaisesRegex = unittest.TestCase.assertRaisesRegex
except:
    assertRaisesRegex = unittest.TestCase.assertRaisesRegexp

## Python-2 compatibility
#
try:
    FileNotFoundError
except NameError:
    FileNotFoundError = IOError  # @ReservedAssignment
else:
    FileNotFoundError = OSError  # @ReservedAssignment

def raise_ex_from(ex_class, chained_ex, *args, **kwds):
 from six import reraise


##############
#  Utilities
#
def str2bool(v):
    vv = v.lower()
    if (vv in ("yes", "true", "on")):
        return True
    if (vv in ("no", "false", "off")):
        return False
    try:
        return float(v)
    except:
        raise argparse.ArgumentTypeError('Invalid boolean(%s)!' % v)


def pairwise(t):
    '''From http://stackoverflow.com/questions/4628290/pairs-from-single-list'''
    it1 = iter(t)
    it2 = iter(t)
    try:
        next(it2)
    except:
        return []
    return zip(it1, it2)


## From http://code.activestate.com/recipes/578231-probably-the-fastest-memoization-decorator-in-the-/
#
def memoize(f):
    """ Memoization decorator for functions taking one or more arguments. """
    class memodict(dict):
        def __init__(self, f):
            self.f = f
        def __call__(self, *args):
            return self[args]
        def __missing__(self, key):
            ret = self[key] = self.f(*key)
            return ret
    return memodict(f)




## From http://stackoverflow.com/a/4149190/548792
#
class Lazy(object):
    def __init__(self,func):
        self.func=func
    def __str__(self):
        return self.func()


def is_travis():
    return 'TRAVIS' in os.environ

def generate_filenames(filename):
    f, e = os.path.splitext(filename)
    yield filename
    i = 1
    while True:
        yield '%s%i%s' % (f, i, e)
        i += 1


def open_file_with_os(fpath):
    ## From http://stackoverflow.com/questions/434597/open-document-with-default-application-in-python
    #     and http://www.dwheeler.com/essays/open-files-urls.html
    import subprocess
    try:
        os.startfile(fpath)
    except AttributeError:
        if sys.platform.startswith('darwin'):
            subprocess.call(('open', fpath))
        elif os.name == 'posix':
            subprocess.call(('xdg-open', fpath))
    return


###### WINDOWS ######
#####################

## From: http://stackoverflow.com/questions/2216173/how-to-get-path-of-start-menus-programs-directory
#
def win_shell():
    from win32com.client import Dispatch
    return Dispatch('WScript.Shell')

def win_folder(wshell, folder_name, folder_csidl=None):
    """

    :param wshell: win32com.client.Dispatch('WScript.Shell')
    :param str folder_name: ( StartMenu | MyDocuments | ... )
    :param str folder_csidl: see http://msdn.microsoft.com/en-us/library/windows/desktop/dd378457(v=vs.85).aspx
    """
    #from win32com.shell import shell, shellcon                          #@UnresolvedImport
    #folderid = operator.attrgetter(folder_csidl)(shellcon)
    #folder = shell.SHGetSpecialFolderPath(0, folderid)
    folder = wshell.SpecialFolders(folder_name)

    return folder

## See: http://stackoverflow.com/questions/17586599/python-create-shortcut-with-arguments
#    http://www.blog.pythonlibrary.org/2010/01/23/using-python-to-create-shortcuts/
#    but keep that for the future:
#        forgot chose: http://timgolden.me.uk/python/win32_how_do_i/create-a-shortcut.html
def win_create_shortcut(wshell, path, target_path, wdir=None, target_args=None, icon_path=None, desc=None):
    """

    :param wshell: win32com.client.Dispatch('WScript.Shell')

    """

    is_url = path.lower().endswith('.url')
    shcut = wshell.CreateShortCut(path)
    try:
        shcut.Targetpath = target_path
        if icon_path:
            shcut.IconLocation = icon_path
        if desc:
            shcut.Description = desc
        if target_args:
            shcut.Arguments = target_args
        if wdir:
            shcut.WorkingDirectory = wdir
    finally:
        shcut.save()

def win_wshell():
    from win32com.client import Dispatch

    shell = Dispatch('WScript.Shell')

    return shell




#!/usr/bin/env python
#-*- coding: utf-8 -*-
#
# Copyright 2013-2019 European Commission (JRC);
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl

import os
import sys
from wltp import (plots, model)
from wltp.test import wltp_db_tests as wltpdb

from matplotlib import pyplot as plt, cm

import pandas as pd


classes = ['class1', 'class2', 'class3b']
part_names = model.get_class_part_names()
def class_name_from_num(cls_num):
    return classes[cls_num-1]

def data_meanN_gears(cls_num):
    """
    Constructs meanN vs gears, hierarchically grouped by: vehicle, python/heinz, class_part
    
    :return: df like that::
                
                                  python                  heinz             
                                    gear            n      gear            n
                1154 Low        2.020374  1078.312480  2.039049  1105.891341
                     Medium     3.577367  1364.552656  3.618938  1353.300231
                     High       4.028571  1694.902241  4.160440  1618.457143
                     ExtraHigh  4.266254  2532.802858  4.424149  2514.894737

    1155 ...
    """
    def means_by_parts(df, cols_in, cols_out):
        df = df.ix[:, cols_in]
        data = df.groupby(pd.cut(df_g.index, bins=parts)).mean()
        data.index = part_names[:len(data.index)]
        data.columns = cols_out
        
        return data

    cls_name = class_name_from_num(cls_num)
    
    vehdata = wltpdb._run_the_experiments(transplant_original_gears=False, compare_results=False)
    vehdata = vehdata[vehdata['class'] == cls_num]    ## Filter only the  Class.
    
    cols_out = ['gear', 'n']
    data = {}
    for (veh_num, df_g, df_h) in wltpdb._file_pairs(wltpdb.gened_fname_glob):
        try:
            class_num = vehdata.loc[veh_num, 'class']
        except KeyError:
            continue
        parts = model.get_class_parts_limits(cls_name, edges=True)
        
        g_data = means_by_parts(df_g, ['gears', 'rpm'], cols_out)
        h_data = means_by_parts(df_h, ['gear', 'n'], cols_out)
        
        veh_data = pd.concat( (g_data, h_data), axis=1, keys=['python', 'heinz'])
        data[veh_num] = veh_data

    df = pd.concat(list(data.values()), axis=0, keys=data.keys())

    return df


def plot(cls_num):
    os.chdir(os.path.join(wltpdb.mydir, wltpdb.samples_dir))

    cls_name = class_name_from_num(cls_num)

    data = data_meanN_gears(cls_num)

    cols = data.keys()
    assert len(cols) == 4, cols
    
    parts = zip(model.get_class_part_names(), [
        dict(data_fmt='1', data_kws=dict(color='#0000ff', markersize=4.8, ), 
            diff_kws=dict(linestyle=':', linewidth=4)
        ),  
        dict(data_fmt='_', data_kws=dict(color='#0033cc', markersize=4.8, fillstyle='none'), 
            diff_kws=dict(linestyle='-.',linewidth=4)
        ),  
        dict(data_fmt='o', data_kws=dict(color='#006699', markersize=4.8, fillstyle='none'), 
            diff_kws=dict(linestyle='-',linewidth=4)
        ), 
        dict(data_fmt='+', data_kws=dict(color='#009966', markersize=4.8, fillstyle='none'), 
            diff_kws=dict(linestyle='--', linewidth=4)
        ), 
    ])
    
    axes_tuple = None
    for (part, kws) in parts:
        try:
                        ## veh         part   N/gear/Py/Heinz
            X = data.loc[(slice(None), part), cols[0]]
            Y = data.loc[(slice(None), part), cols[1]]
            X_REF = data.loc[(slice(None), part), cols[2]]
            Y_REF = data.loc[(slice(None), part), cols[3]]
        except KeyError:
            continue
        axes_tuple, artist_tuple = plots.plot_xy_diffs_arrows(
            X, Y, X_REF, Y_REF,
            data_label='Python, %s'%part, diff_label='Diffs %s'%part,
            diff_cmap=cm.cool, #@UndefinedVariable
            title="%s: Python(arrow-heads) compared to AccessDb-2secAccel (arrow-tails)" % cls_name,
            x_label=r'Mean Gear',
            y_label='$Mean EngineSpeed [rpm]$',
            axes_tuple=axes_tuple,
            **kws
        )

    axes_tuple[0].legend(loc=1, fancybox=True, framealpha=0.5)
    axes_tuple[1].legend(loc=2, fancybox=True, framealpha=0.5)

    return data, axes_tuple


if __name__ == '__main__':
    cls_num = int(sys.argv[1]) if len(sys.argv) > 1 else 2
    data, axes_tuple = plot(cls_num)
    
    ## Add Pick-cursor
    #
    main_axes = axes_tuple[0]
    # Workaround Matplotlib#815: pick_event works only on z-top-axes.
    for ax in axes_tuple[1:]: ax.set_zorder(0.1); main_axes.set_zorder(1)
    fig = main_axes.get_figure()
    cursor = plots.DataCursor(
                              ax=main_axes, 
                              annotations=list(data.index.levels[0]),
                              text_template='Vehicle: {txt}')
    fig.canvas.mpl_connect('pick_event', cursor)

    plt.show()




#!/usr/bin/env python
#-*- coding: utf-8 -*-
#
# Copyright 2013-2019 European Commission (JRC);
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl
'''Plot the class3 diagram used in the documents and stores in project's ``docs/_static`` folder.

Invoked with 1 arg will draw diagram on screen and wait before ending.
Invoked with 2 args will additionally save image as well.
'''

from __future__ import division, print_function, unicode_literals

import sys, os, os.path as path

from matplotlib import pyplot as plt
from matplotlib.pyplot import pause
from wltp import model

import numpy as np


def make_class_fig(class_name, class_data):
    font_size = 8
    tick_size = 6.5

    ## Get cycle-data
    #
    c=np.array(class_data['cycle'])
    a = 1000 * np.gradient(c) / 3600
    t = np.arange(0, len(c))
    parts = class_data['parts']
    part_limits = [e+0.5 for (s, e) in parts[:-1]]

    fig, ax1 = plt.subplots()
    ax1.set_xlabel('Time (s): ', size=font_size)
    ax1.xaxis.set_label_coords(-0.06, -0.030)
    for tl in ax1.get_xticklabels():
        tl.set_size(tick_size)


    ax1.set_ylabel(r'Velocity ($km/h$)', color='b', size=font_size)
    for tl in ax1.get_yticklabels():
        tl.set_color('b')
        tl.set_size(tick_size)


    ax2 = ax1.twinx()
    ax2.set_ylabel(r'Acceleration ($m/\sec_2$)', color='m', size=font_size)
    for tl in ax2.get_yticklabels():
        tl.set_color('m')
        tl.set_size(tick_size)

    l_vel = ax2.plot(t, a, 'm-', alpha=0.3)[0]
    l_acc = ax1.plot(t, c, 'b-')[0]
#     plt.legend([l_vel, l_acc], ['Velocity', 'Acceleration'])

    plt.title('WLTC %s'%class_name,  size=9.5, weight='bold')

    ## Plot part-Limits
    #
    for limit in part_limits:
        l = plt.axvline(limit, color='r', linewidth=2)

    ## Add part-labels.
    #
    v_pos = 129.5 # trial'n error
    if class_name == 'class1': # Acceleration scale changes!!
        v_pos /= 2
    bbox={'facecolor':'red', 'alpha':0.5, 'pad':4, 'linewidth':0}
    txts = ['Low', 'Medium', 'High', 'ExtraHigh']
    txts_pos = [0] + part_limits #[0.40, 0.67, 0.85]

    for (txt, h_pos) in zip(txts, txts_pos):
        ax1.text(h_pos + 8, v_pos, txt, style='italic',
            bbox=bbox, size=8)


    ax1.grid()
    ax1.xaxis.grid = True
    ax1.yaxis.grid = True

    fig.set_size_inches(2 * 2 * 1.618, 2)

    return fig


if __name__ == '__main__':
    os.chdir(path.dirname(__file__))

    wltc_data = model._get_wltc_data()
    for (class_name, class_data) in wltc_data['classes'].items():
        fig = make_class_fig(class_name, class_data)

        img_fname = path.join('..', 'docs', '_static', 'wltc_%s.png'%class_name )

        nargs = len(sys.argv)
        if (nargs > 1):
            plt.show()
        if (nargs == 1 or nargs > 2):
            fig.savefig(img_fname, dpi=100)
            print('Stored image: %s'% img_fname)




#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2014 European Commission (JRC);
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl
'''
Detects vehile's gear-ratios from cycle-data and reconstructs the gears-profile by identifying the actual gears used.

The following terms and arguments are used throughout this module:

gear-ratio(s):
    the "NoverV", which is the inverse of the "engine-speed-to-velocity" ("STV").
    Conceptually it is the number of engine revolutions required, in RPM,
    in order for the the vehicle to travel with 1 km/h.

    .. Note: The inverse of STVs ("VoverN") are used throughout the calculations because
            they are practically between (0, 1), and more equally spaced.

cycle-ratios:
    The actual or detected NoverV-ratio on each cycle-point.

cycle-gears:
    The actual or identified gear used on each cycle-point.

cycle-data
    a :class:`pd.DataFrame` containing (at the begining) `N` and `V` columns
    with [rpm] and [km/h] as units respectively.

The detection of the gear-ratios happens in a 3 steps, approximation or guessing, estimation and selection:

1. Produce a number of different *approximate* sets of gear-ratios (the "guesses")
   by producing histograms of the ratios with different combinations of bins, and then
   taking the first #gears of peaks.
2. Feed these "guessed" sets of gear-ratios into a robust (with median) kmeans clustering algorithm, and
3. pick the gear-ratios with the minimum distortion.

The identification of the actual gear on each cycle-data point (reconstructing the gear-profile)
happens in 2 steps:

1. On each cycle-point select the Gear with smallest absolute-difference with the respective identified ratio;
2. Reconstruct engine-speed by dividing velocity-profile with identified gear-ratio on each cycle-point.

'''
## TODO: Rename columns N, V.
## TODO: Classify code.

from collections import namedtuple
import functools as ft
import math
import numpy as np, pandas as pd
from scipy import signal


def dequantize_unstabbleMean(df, method='linear'):
    """
    :param str method: see :method:`pd.DataFrame.interpolate()` that is used to fill-in column-values that are held constant
    """
    edges       = df.diff()

    df_1        = df.where(edges != 0)              ## Points after change.
    df_0        = df.where((edges.shift(-1) != 0))  ## Points before change

    df_21       = 2 * (df_0.shift(1) + df_1) / 3    ## 2/3 * sum located after change
    df_12       = df_21.shift(-1) / 2               ## 1/3 * sum located before change
    df2         = df_21 + df_12

    df2.iloc[[0, -1], :] = df.iloc[[0, -1], :] ## Pin start-end values to originals.

    df2.interpolate(method=method, inplace=True)

    return df2


def dequantize(df, method='linear'):
    """
    Tries to undo results of a non-constant sampling rate (ie due to indeterministic buffering).

    :param str method: see :method:`pd.DataFrame.interpolate()` that is used to fill-in column-values that are held constant

    .. Note::
        The results wil never have a flat section, so checking for some value (ie 0)
        must always apply for some margin (ie (-e-4, e-4).
    """
    df1 = df.where(df.diff() != 0)
    df1.iloc[[0, -1]] = df.iloc[[0, -1]] ## Pin start-end values to originals.

    df1.interpolate(method=method, inplace=True)

    return df1


def identify_n_idle(N):
    ## Estimate n_idel as the peak-N on the lower-half.
    #
    N = N[N < N.median()]
    peaks, _, _ = find_histogram_peaks(N, math.sqrt(len(N)))

    n_idle_0 = peaks.iloc[0, -1]
    #print(n_idle_0)

    ## Rebin-smoothly around 1-stdev of estimated n_idle,
    #    to increase accuracy.
    #
    jog = N.std() / 2
    N = N[(N > (n_idle_0 - jog)) & (N < (n_idle_0 + jog))]
    peaks, _, _ = find_histogram_peaks(N, math.sqrt(len(N)), 3)

    return peaks.iloc[0, -1]


def _outliers_filter_df2(df, cols):
    for col in cols:
        mean = df[col].median()
        #std  = np.subtract(*np.percentile(df[col], [75, 25]))
        std  = df[col].mad()
        df = df[
            (df[col] >= mean - 2 * std) &
            (df[col] <= mean + 2 * std)
        ]

    return df

def _outliers_filter_df(df, cols):
    def filter_col(col):

        mean = df[col].median()
        #std  = np.subtract(*np.percentile(df[col], [75, 25]))
        std  = df[col].mad()
        return np.abs(df[col] - mean) <= 2 * std

    return ft.reduce((filter_col(col) for col in cols), np.bitwise_and)

def _transient_filter_df(df, col, std_threshold=0.2, apply_dequantize=True):
    if apply_dequantize:
        df[col] = dequantize(df[col])
    grad = np.gradient(df[col])
    grad_mean, grad_std = grad.mean(), grad.std()
    thres = (
        grad_mean - std_threshold * grad_std,
        grad_mean + std_threshold * grad_std
    )
    df1 = df.ix[(thres[0] < grad) & (grad < thres[1]), :]

    return df1

def _n_idle_filter_df(df, col, n_idle_threshold=1.05):
    n_idle = identify_n_idle(df[col])

    df1 = df[df[col] > n_idle_threshold * n_idle]

    return df1


def _filter_cycle_and_derive_ratios(df, filter_outliers=None, filter_transient=True):
    df['VoverN'] = (df.V / df.N) # Work with VoverN because evenly spaced (not log), and 0 < VoverN < 1
    df0 = df[(df.V > 2) & (df.VoverN > 0)]
    df1 = _n_idle_filter_df(df0, 'N')
    df2 = _transient_filter_df(df1, 'VoverN')
#     valids = _outliers_filter_df(df2, ['VoverN']) if filter_outliers else df2 # PROBLEMATIC!!
#     df3 = df2[valids]

    return df2


def moving_average(x, window):
    """
    :param int window: odd windows preserve phase

    From http://nbviewer.ipython.org/github/demotu/BMC/blob/master/notebooks/DataFiltering.ipynb
    """
    return np.convolve(x, np.ones(window)/window, 'same')


def find_histogram_peaks(df, bins, smooth_window=None):
    h_points, bin_edges = np.histogram(df, bins=bins, density=True)
    h_centers = (bin_edges[:-1] + bin_edges[1:]) / 2

    if smooth_window:
        h_points = moving_average(h_points, smooth_window)
    peak_bins = signal.argrelmax(h_points, mode='clip')
    peak_bins = peak_bins[0]

    peaks = pd.DataFrame({'bin': peak_bins, 'value': h_centers[peak_bins], 'population': h_points[peak_bins]})
    peaks = peaks.sort(['population'], ascending=False)

    return peaks, h_centers, h_points


def _approximate_ratios_by_binning(ratios, bins, ngears):
    """
    :return: 3 sets of ratios:
                1. all peaks detected with bins specified
                2. the first #gears peaks
                3. the first #gears peaks detected by binning the space between the previous #gears identified
    """
    peaks0, _, _ = find_histogram_peaks(ratios, bins)
    #display(peaks0)

    peaks1 = peaks0[:ngears].sort(['value'], ascending=True) ## Keep #<gears> of most-populated bins.
    #display(peaks1)

    ## Re-bin with narrower X-to-bin.
    #
    r_min = peaks1['value'].min()
    r_max = peaks1['value'].max()
    gear_distance = (r_max - r_min) / ngears
    r_min -= gear_distance
    r_max += gear_distance
    peaks2, h_centers, h_points = find_histogram_peaks(ratios[(ratios >= r_min) & (ratios <= r_max)], bins)
    #display(peaks2)

    peaks3 = peaks2[:ngears].sort(['value'], ascending=True) ## Keep #<gears> of most-populated bins.
    #display(peaks3)

    return peaks0, peaks1, peaks3, h_points, h_centers


def _norm1_1d_vq(obs, guess, is_robust=False):
    """Simplified from scipy.cluster.vq.py_vq2()"""
    dist = np.abs(obs[np.newaxis, :] - guess[:, np.newaxis])
    if is_robust: ## Does not make a difference!
        robust_prcntile = 4.685     ##  Bisquare M-estimator with 95% efficiency under the Gaussian model.
        R_deleved   = dist / (robust_prcntile * 1.4826 * np.median(dist))  ## * sqrt(1 - hat_vector))
        ## Calc the robust bisquared-residuals excluding outliers.
        R_weights   = (R_deleved < 1) * (1 - R_deleved**2)**2
        dist = R_weights * dist

    code = np.argmin(dist, 0)
    min_dist = np.minimum.reduce(dist, 0)

    return code, min_dist

def _estimate_ratios_by_kmeans(obs, guess, **kmeans_kws):
    """
    Adapted kmeans to use norm1 distances (identical to euclidean-norm2 for 1D)
    and using median() for each guess, to ignore outliers and identify
    "gear-lines".
    """

    def _1d_kmeans(obs, guess, guess_func, thresh=1e-5):
        """ The "raw" version of k-means, adapted from scipy.cluster.vq._kmeans(). """

        code_book = np.array(guess, copy=True)
        avg_dist = []
        diff = thresh+1.
        while diff > thresh:
            nc = code_book.size
            # compute membership and distances between obs and code_book
            obs_code, distort = _norm1_1d_vq(obs, code_book)
            avg_dist.append(np.mean(distort))
            # recalc code_book as centroids of associated obs
            if(diff > thresh):
                has_members = []
                for i in np.arange(nc):
                    cell_members = obs[obs_code == i]
                    if cell_members.size > 0:
                        code_book[i] = guess_func(cell_members)
                        has_members.append(i)
                # remove code_books that didn't have any members
                code_book = code_book[has_members]
            if len(avg_dist) > 1:
                diff = avg_dist[-2] - avg_dist[-1]
        # print avg_dist
        return code_book, avg_dist[-1]

    centers, distortion = _1d_kmeans(obs, guess, guess_func=np.median, **kmeans_kws)

    centers = np.sort(centers)

    return centers, distortion


## The artifacts of each detection attempt.
#
Detekt = namedtuple('Detekt', [
    'distort',
    'guess', 'final',        ##
    'all_peaks_df',
    'hist_X', 'hist_Y'
])
def _new_Detekt_approximation(guess_ratios, all_peaks_df, hist_X, hist_Y):
    return Detekt(None, guess_ratios, None, all_peaks_df, hist_X, hist_Y)
def _set_Detekt_final(gears, final_ratios, distortion):
    return gears._replace(distort=distortion, final=final_ratios)

def _append_aproximate_Detekt(ngears, cases, ratios, all_peaks_df, hist_X, hist_Y):
    """Appends 2 sets of ratios for each one specified as input, the 2nd one being a linspace between the min-max."""
    if ratios.size == ngears:
        cases.append(_new_Detekt_approximation(ratios, all_peaks_df, hist_X, hist_Y))
        cases.append(_new_Detekt_approximation(np.linspace(ratios.min(), ratios.max(), ngears), all_peaks_df, hist_X, hist_Y))

def _gather_guessed_Detekts(ngears, cycle_df):
    """Makes a number gear-ratios guesses based on histograms with different bin combinations. """

    guessed_detekts = [] ## A list-of-Detekt:
    bins_cases = [
            (ngears+2) * 6, # Roughly 4 bins per gear plus 2 left & right.
            math.sqrt(cycle_df.shape[0]),
    ]
    for bins in bins_cases:
        all_peaks_df, peaks_df1, peaks_df2, hist_X, hist_Y = _approximate_ratios_by_binning(cycle_df.VoverN, bins=bins, ngears=ngears)

        _append_aproximate_Detekt(ngears, guessed_detekts, peaks_df1.value.values, all_peaks_df, hist_X, hist_Y)
        _append_aproximate_Detekt(ngears, guessed_detekts, peaks_df2.value.values, all_peaks_df, hist_X, hist_Y)

    ## Add linspace for the complete ratio-range
    #    using the 1st Detekt as template.
    guessed_detekts.append(guessed_detekts[0]._replace(guess=np.linspace(cycle_df.VoverN.min(), cycle_df.VoverN.max(), ngears)))

    return guessed_detekts


def _gather_final_Detekts(ngears, cycle_df, guessed_gears):
    """
    :param pd.DataFrame cycle_df: a dataframe with `N` and `V` columns with units [rpm] and [km/h] respectively
                        (or inferred by comparing the means of these 2 columns).
    """
    final_detekts = []
    for gears in guessed_gears:
        ratios, distort = _estimate_ratios_by_kmeans(cycle_df.VoverN.values, gears.guess)
        if ratios.size == ngears:
            final_detekts.append(_set_Detekt_final(gears, ratios, distort))
    final_detekts = sorted(final_detekts, key=lambda x: x.distort)

    return final_detekts


def run_gear_ratios_detections_on_cycle_data(ngears, cycle_df):
    """
    Invoke this one if you want to draw the results.

    :param pd.DataFrame cycle_df: it must contain (at least) `N` and `V` columns (units: [rpm] and [km/h] respectively)
    :return: a list of all :class:`Detekt` tuples sorted with the most probable one at the the head,
                        needed.  Its 1st element is the solution
    """
    filtered_df     = _filter_cycle_and_derive_ratios(cycle_df)
    guessed_detekts = _gather_guessed_Detekts(ngears, filtered_df)
    final_detekts   = _gather_final_Detekts(ngears, filtered_df, guessed_detekts)

    return final_detekts

def detect_gear_ratios_from_cycle_data(ngears, cycle_df):
    """
    Use a 2 step procedure if you want to plot the results, by invoking
    `run_gear_ratios_detections_on_cycle_data()` and `plot_idgears_results()` separately.

    :return: a :class:`ndarray` with the detected gear-ratios (for the STVs, inverse them)
    """
    detekts = run_gear_ratios_detections_on_cycle_data(ngears, cycle_df)
    if not detekts[0].final is None:
        return detekts[0].final, detekts[0].distort
    else:
        raise Exception('Detection failed to estimate any gear-ratios!\n  All-Detekts(%s)' % detekts)


def identify_gears_from_cycle_data(cycle_df, gear_ratios):
    """
    Like :meth:`identify_gears_from_ratios()` but with more sane filtering (ie v above 1 km/h).

    :param pd.DataFrame cycle_df: it must contain (at least) `N` and `V` columns (units: [rpm] and [km/h] respectively)
    :param ndarray gear_ratios: the same as :meth:`identify_gears_from_ratios()`
    :return: the same as :meth:`identify_gears_from_ratios()`
    """
    df = _filter_cycle_and_derive_ratios(cycle_df)
    return identify_gears_from_ratios(df.VoverN, gear_ratios)


def identify_gears_from_ratios(cycle_ratios, gear_ratios):
    """
    Return arrays will miss NaNs!

    :param ndarray cycle_ratios: a single column array/list/ndarray  of ratios (STVs or inverse).
    :param ndarray gear_ratios: this list/ndarray of gear-ratios with len equal to the #gears
    :return: a 2-tuple, where [0] is the 0-based identified-gears, and
                        [1] the distortion for each cycle-point.
    """
    cycle_ratios = np.asarray(cycle_ratios).flatten()
    nums = np.isfinite(cycle_ratios)
    gears, distort = _norm1_1d_vq(cycle_ratios[nums], np.asarray(gear_ratios).flatten())

    return gears, distort

def reconstruct_engine_speed(cycle_df, gear_ratios):
    """
    Adds

    :param ndarray gear_ratios: this list/ndarray of gear-ratios with len equal to the #gears

    TODO: Class-method
    """
    cycle_df['GEARS_detekted'] = identify_gears_from_ratios(cycle_df['VoverN'], gear_ratios) + 1
    cycle_df['VoverN_reconstructed'] = cycle_df['V'] / cycle_df['VoverN']
    cycle_df['N_reconstructed'] = cycle_df['V'] / cycle_df['VoverN']



def plot_idgears_results(cycle_df, detekt, fig=None, axes=None, original_points=None):
    """
    :param detekt: A Detekt-namedtuple with the data to plot
    """
    from matplotlib import pyplot as plt

    if not fig:
        fig = plt.figure(figsize=(18,10))
        fig.suptitle('Detekt: %s, Accuracy: %s (wltp-good: ~< 1e-3)'%(detekt.final, detekt.distort))

    if axes:
        ax1, ax2, ax3 = axes
    else:
        ax1 = plt.subplot(2,2,1)
        ax2 = plt.subplot(2,2,2)
        ax3 = plt.subplot(2,1,2)
        ax31 = ax3.twinx()
        ax32 = ax31.twinx()

    peaks_stv = (1/detekt.guess)
    kmeans_stv = (1/detekt.final)

    ## Plot engine-points
    ##
    ax1.plot(cycle_df.V, cycle_df.N, 'g+', markersize=5)
    if not original_points is None:
        ax1.plot(original_points.V, original_points.N, 'k.', markersize=1)
    for r in peaks_stv:
        ax1.plot([0, cycle_df.V.max()], [0, cycle_df.V.max()*r], 'c-', alpha=0.5)
    for r in kmeans_stv:
        ax1.plot([0, cycle_df.V.max()], [0, cycle_df.V.max()*r], 'r-', alpha=0.5)
    ax1.set_ylim(0, cycle_df.N.median() + 2*cycle_df.N.mad())
    ax1.set_xlim(cycle_df.V.median() - 2*cycle_df.V.mad(), cycle_df.V.median() + 2*cycle_df.V.mad())


    ## Plot ratios's Histogram
    #
    ax2.plot(detekt.hist_Y, detekt.hist_X)
    ax2.plot(detekt.all_peaks_df.value, detekt.all_peaks_df.population, 'ob', markersize=8, fillstyle='none')
    #ax2.plot(peaks_df.value, peaks_df.population, 'ob', markersize=8, fillstyle='full') ## Annotate top-#detekt
    # plt.hlines(detekt.all_peaks_df.population.mean() - detekt.all_peaks_df.population.mad(), 0, detekt.hist_Y.max(), 'g', linestyle='-')

    ## Scatter-plot Ratios
    ##
    R = cycle_df.VoverN
#     ax31.plot(cycle_df.N, 'm-', lw=0.5, markersize=2)
    ax31.plot(cycle_df.V, 'b-', lw=0.5, markersize=2)
    ax3.plot(R, 'g.-', lw=0.5, markersize=2)
    if not original_points is None:
        ax3.plot(original_points.V / original_points.N, 'k.', markersize=1)
    for r in detekt.all_peaks_df.value:
        plt.hlines(r, 0, R.shape[0], 'c', linestyle=':')
    for r in detekt.final:
        plt.hlines(r, 0, R.shape[0], 'c', linestyle='-.')
    for r in detekt.final:
        plt.hlines(r, 0, R.shape[0], 'r', linestyle='--')
    plt.ylim(R.min(), R.max())
    plt.xlim(0, R.shape[0])





#!/usr/bin/env python
#-*- coding: utf-8 -*-
#
# Copyright 2013-2019 European Commission (JRC);
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl
'''
wltp: generate WLTC gear-shifts based on vehicle characteristics
'''

from __future__ import division, unicode_literals

from ._version import (__version__, __updated__)


__copyright__     = "Copyright (C) 2013-2014 European Commission (JRC)"  # noqa
__license__       = "EUPL 1.1+"                     # noqa
__title__     = "wltp"                              # noqa
__summary__   = __doc__.splitlines()[0]             # noqa
__uri__       = "https://github.com/JRCSTU/wltp"  # noqa
